from winacl.dtyp.ace import ACE
import io
import enum
import hashlib
from typing import List

class ACL_REVISION(enum.Enum):
	NO_DS = 0x02 # When set to 0x02, only AceTypes 0x00, 0x01, 0x02, 0x03, 0x11, 0x12, and 0x13 can be present in the ACL. An AceType of 0x11 is used for SACLs but not for DACLs. For more information about ACE types, see section 2.4.4.1.
	DS = 0x04 #When set to 0x04, AceTypes 0x05, 0x06, 0x07, 0x08, and 0x11 are allowed. ACLs of revision 0x04 are applicable only to directory service objects. An AceType of 0x11 is used for SACLs but not for DACLs.

ACL_REV_NODS_ALLOWED_TYPES = [0x00, 0x01, 0x02, 0x03, 0x11, 0x12, 0x13]
ACL_REV_DS_ALLOWED_TYPES   = [0x05, 0x06, 0x07, 0x08, 0x11]

class ACL:
	def __init__(self, sd_object_type = None):
		self.AclRevision:int = None
		self.Sbz1:int = 0
		self.AclSize:int = None
		self.AceCount:int = None
		self.Sbz2:int = 0
		
		self.aces:List[ACE] = []
		self.sd_object_type = sd_object_type
		
	@staticmethod
	def from_buffer(buff, sd_object_type = None):
		acl = ACL(sd_object_type)
		acl.AclRevision = int.from_bytes(buff.read(1), 'little', signed = False)
		acl.Sbz1 = int.from_bytes(buff.read(1), 'little', signed = False)
		acl.AclSize = int.from_bytes(buff.read(2), 'little', signed = False)
		acl.AceCount = int.from_bytes(buff.read(2), 'little', signed = False)
		acl.Sbz2 = int.from_bytes(buff.read(2), 'little', signed = False)
		for _ in range(acl.AceCount):
			acl.aces.append(ACE.from_buffer(buff, sd_object_type))
		return acl

	def to_bytes(self):
		buff = io.BytesIO()
		self.to_buffer(buff)
		buff.seek(0)
		return buff.read()

	def to_buffer(self, buff):
		data_buff = io.BytesIO()

		self.AceCount = len(self.aces)
		for ace in self.aces:
			ace.to_buffer(data_buff)

		self.AclSize = 8 + data_buff.tell()

		buff.write(self.AclRevision.to_bytes(1, 'little', signed = False))
		buff.write(self.Sbz1.to_bytes(1, 'little', signed = False))
		buff.write(self.AclSize.to_bytes(2, 'little', signed = False))
		buff.write(self.AceCount.to_bytes(2, 'little', signed = False))
		buff.write(self.Sbz2.to_bytes(2, 'little', signed = False))
		data_buff.seek(0)
		buff.write(data_buff.read())
		
	def __str__(self):
		t = '=== ACL ===\r\n'
		for ace in self.aces:
			t += '%s\r\n' % str(ace)
		return t

	def to_sddl(self, object_type = None):
		t = ''
		for ace in self.aces:
			t += ace.to_sddl(object_type)
		return t

	@staticmethod
	def from_sddl(sddl_str, object_type = None, domain_sid = None):
		acl = ACL()
		acl.AclRevision = 2
		acl.AceCount = 0
		
		for ace_sddl in sddl_str.split(')('):
			ace = ACE.from_sddl(ace_sddl, object_type = object_type, domain_sid = domain_sid)
			acl.aces.append(ace)
			acl.AceCount += 1
			if acl.AclRevision == 2:
				if ace.AceType.value in ACL_REV_DS_ALLOWED_TYPES:
					acl.AclRevision = ACL_REVISION.DS.value

		return acl

	def __eq__(self, acl):
		if not isinstance(acl, ACL):
			return False

		if self.AclRevision != acl.AclRevision:
			return False
		if self.Sbz1 != acl.Sbz1:
			return False
		if self.Sbz2 != acl.Sbz2:
			return False
		this_aces = {}
		that_aces = {}
		for ace in self.aces:
			buff = io.BytesIO()
			ace.to_buffer(buff)
			pos = buff.tell()
			buff.seek(0,0)
			this_aces[hashlib.sha1(buff.read(pos)).digest()] = 1
		
		for ace in acl.aces:
			buff = io.BytesIO()
			ace.to_buffer(buff)
			pos = buff.tell()
			buff.seek(0,0)
			that_aces[hashlib.sha1(buff.read(pos)).digest()] = 1

		for ha in this_aces:
			if ha not in that_aces:
				return False
		
		for ha in that_aces:
			if ha not in this_aces:
				return False

		for hthis, hthat in zip(this_aces.keys(), that_aces.keys()):
			if hthis != hthat:
				return False

		return True

	def diff(self, acl):
		diff_res = {}
		if self.AclRevision != acl.AclRevision:
			diff_res['revision'] = [self.AclRevision, acl.AclRevision]
		if self.Sbz1 != acl.Sbz1:
			diff_res['Sbz1'] = [self.Sbz1, acl.Sbz1]
		if self.Sbz2 != acl.Sbz2:
			diff_res['Sbz2'] = [self.Sbz2, acl.Sbz2]
		this_aces = {}
		that_aces = {}
		for ace in self.aces:
			buff = io.BytesIO()
			ace.to_buffer(buff)
			pos = buff.tell()
			buff.seek(0,0)
			data = buff.read(pos)
			this_aces[hashlib.sha1(data).digest()] = data
		
		for ace in acl.aces:
			buff = io.BytesIO()
			ace.to_buffer(buff)
			pos = buff.tell()
			buff.seek(0,0)
			data = buff.read(pos)
			that_aces[hashlib.sha1(data).digest()] = data

		for ha in this_aces:
			if ha not in that_aces:
				diff_res['deleted'] = this_aces[ha]
		
		for ha in that_aces:
			if ha not in this_aces:
				diff_res['added'] = that_aces[ha]

		#TODO: diff the ordering changes of ACEs
		#for hthis, hthat in zip(this_aces.keys(), that_aces.keys()):
		#	if hthis != hthat:
		#		return False

		return diff_res


from .defines import *

GENERIC_READ                     = 0x80000000
GENERIC_WRITE                    = 0x40000000
GENERIC_EXECUTE                  = 0x20000000
GENERIC_ALL                      = 0x10000000
READ_CONTROL                     = 0x00020000

FILE_SHARE_READ                  = 0x00000001
FILE_SHARE_WRITE                 = 0x00000002
FILE_SHARE_DELETE                = 0x00000004

CREATE_NEW                       = 1
CREATE_ALWAYS                    = 2
OPEN_EXISTING                    = 3
OPEN_ALWAYS                      = 4
TRUNCATE_EXISTING                = 5

FILE_ATTRIBUTE_READONLY          = 0x00000001
FILE_ATTRIBUTE_NORMAL            = 0x00000080
FILE_ATTRIBUTE_TEMPORARY         = 0x00000100

FILE_FLAG_WRITE_THROUGH          = 0x80000000
FILE_FLAG_NO_BUFFERING           = 0x20000000
FILE_FLAG_RANDOM_ACCESS          = 0x10000000
FILE_FLAG_SEQUENTIAL_SCAN        = 0x08000000
FILE_FLAG_DELETE_ON_CLOSE        = 0x04000000
FILE_FLAG_OVERLAPPED             = 0x40000000

FILE_FLAG_BACKUP_SEMANTICS       = 0x02000000


FILE_ATTRIBUTE_READONLY          = 0x00000001
FILE_ATTRIBUTE_HIDDEN            = 0x00000002
FILE_ATTRIBUTE_SYSTEM            = 0x00000004
FILE_ATTRIBUTE_DIRECTORY         = 0x00000010
FILE_ATTRIBUTE_ARCHIVE           = 0x00000020
FILE_ATTRIBUTE_DEVICE            = 0x00000040
FILE_ATTRIBUTE_NORMAL            = 0x00000080
FILE_ATTRIBUTE_TEMPORARY         = 0x00000100

# DWORD WINAPI GetLastError(void);
def GetLastError():
    _GetLastError = windll.kernel32.GetLastError
    _GetLastError.argtypes = []
    _GetLastError.restype  = DWORD
    return _GetLastError()

# HLOCAL WINAPI LocalFree(
#   __in  HLOCAL hMem
# );
def LocalFree(hMem):
    _LocalFree = windll.kernel32.LocalFree
    _LocalFree.argtypes = [HLOCAL]
    _LocalFree.restype  = HLOCAL

    result = _LocalFree(hMem)
    if result != NULL:
        ctypes.WinError()

# BOOL WINAPI CloseHandle(
#   __in  HANDLE hObject
# );
def CloseHandle(hHandle):
    _CloseHandle = windll.kernel32.CloseHandle
    _CloseHandle.argtypes = [HANDLE]
    _CloseHandle.restype  = bool
    _CloseHandle.errcheck = RaiseIfZero
    _CloseHandle(hHandle)

def CreateFileW(lpFileName, dwDesiredAccess = GENERIC_ALL, dwShareMode = 0, lpSecurityAttributes = None, dwCreationDisposition = OPEN_ALWAYS, dwFlagsAndAttributes = FILE_ATTRIBUTE_NORMAL, hTemplateFile = None):
    _CreateFileW = windll.kernel32.CreateFileW
    _CreateFileW.argtypes = [LPWSTR, DWORD, DWORD, LPVOID, DWORD, DWORD, HANDLE]
    _CreateFileW.restype  = HANDLE

    if not lpFileName:
        lpFileName = None
    if lpSecurityAttributes:
        lpSecurityAttributes = ctypes.pointer(lpSecurityAttributes)
    hFile = _CreateFileW(lpFileName, dwDesiredAccess, dwShareMode, lpSecurityAttributes, dwCreationDisposition, dwFlagsAndAttributes, hTemplateFile)
    if hFile == INVALID_HANDLE_VALUE:
        raise ctypes.WinError()
    return hFile


import io

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.backends import default_backend

class ECDHPrivateKeyBlob:
	def __init__(self):
		self.magic = None
		self.length = None
		self.x = None
		self.y = None
		self.privateexp = None

	@staticmethod
	def from_bytes(data):
		return ECDHPrivateKeyBlob.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		pk = ECDHPrivateKeyBlob()
		pk.magic = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pk.length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pk.x = int.from_bytes(buff.read(pk.length), byteorder='little', signed=False)
		pk.y = int.from_bytes(buff.read(pk.length), byteorder='little', signed=False)
		pk.privateexp = int.from_bytes(buff.read(pk.length), byteorder='little', signed=False)
		return pk
	
	def get_curve(self):
		if self.magic in [0x314B4345, 0x324B4345, 0x31534345, 0x32534345]:
			return ec.SECP256R1()
		elif self.magic in [0x334B4345, 0x344B4345, 0x33534345, 0x34534345]:
			return ec.SECP384R1()
		elif self.magic in [0x354B4345, 0x364B4345, 0x35534345, 0x36534345]:
			return ec.SECP521R1()
		else:
			raise Exception('Unknown curve type for %s' % self.magic.to_bytes(4, byteorder='little', signed=False).hex())
	
	def get_key(self):
		curve = self.get_curve()
		
		private_numbers = ec.EllipticCurvePrivateNumbers(
			private_value=self.privateexp,
			public_numbers=ec.EllipticCurvePublicNumbers(
				x=self.x,
				y=self.y,
				curve=curve
			)
		)
		private_key = private_numbers.private_key(default_backend())
		return private_key
	
	def get_pubkey(self):
		curve = self.get_curve()
		
		public_numbers = ec.EllipticCurvePublicNumbers(
			x=self.x,
			y=self.y,
			curve=curve
		)
		public_key = public_numbers.public_key(default_backend())
		return public_key


	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
		return t


from winacl.dtyp.ace import ACEType, AceFlags
from winacl.dtyp.sid import SID
from typing import List

SYNCHRONIZE = 0x00100000
WRITE_OWNER = 0x00080000
WRITE_DACL = 0x00040000
READ_CONTROL = 0x00020000
DELETE = 0x00010000
MAXIMUM_ALLOWED = 0x02000000

def group_lookup(sid, sid_groups):
	for csid in sid_groups:
		if sid == csid:
			return True
	return False

def sid_in_dacl(sid, dacl):
	for ace in dacl.aces:
		if ace.Sid == sid:
			return True
	return False

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/4f1bbcbb-814a-4c70-a11e-2a5b8779a6f9
def EvaluateSidAgainstDescriptor(sd, sid:str or SID, req_access:int, sid_groups:List[str] = []):
	# sid_groups is a list of SIDs where the user's SID is member of
	dacl = sd.Dacl
	sacl = sd.Sacl
	remainging_access = req_access
	allowed_access = 0
	denied_access = 0
	max_allowed_mode = False
	granted_access = 0

	#IF RemainingAccess contains ACCESS_SYSTEM_SECURITY access bit THEN
	#     IF Token.Privileges contains SeSecurityPrivilege THEN
	#         Remove ACCESS_SYSTEM_SECURITY access bit from RemainingAccess
	#         Set GrantedAccess to GrantedAccess or ACCESS_SYSTEM_SECURITY
	#
	#         IF RemainingAccess to 0 THEN
	#             Return success
	#         Else
	#
	#     ELSE
	#         Set GrantedAccess to 0
	#         Return access_denied
	#
	#     END IF
	# END IF

	#IF RemainingAccess contains WRITE_OWNER access bit and Token.Privileges is not NULL THEN
	#     IF  Token.Privileges contains SeTakeOwnershipPrivilege THEN
	#       Remove WRITE_OWNER access bit from RemainingAccess
	#       Set GrantedAccess to GrantedAccess or WRITE_OWNER
	#    END IF
	# END IF

	#if WRITE_OWNER in remainging_access:
	#    remainging_access = WRITE_OWNER
	#    granted_access |= WRITE_OWNER

	#-- the owner of an object is always granted READ_CONTROL and WRITE_DAC.
	# CALL SidInToken(Token, SecurityDescriptor.Owner, PrincipalSelfSubst)
	# IF SidInToken returns True THEN
	#     IF DACL does not contain ACEs from object owner THEN
	#         Remove READ_CONTROL and WRITE_DAC from RemainingAccess
	#         Set GrantedAccess to GrantedAccess or READ_CONTROL or WRITE_OWNER
	#     END IF
	# END IF

	if group_lookup(sd.Owner, sid_groups) is True:
		if sid_in_dacl(sd.Owner, dacl) is False:
			remainging_access &= ~READ_CONTROL
			remainging_access &= ~WRITE_DACL
			granted_access |= WRITE_OWNER 
			granted_access |= READ_CONTROL
	
	if bool(remainging_access & MAXIMUM_ALLOWED) is True :
		max_allowed_mode = True

	for ace in dacl.aces:
		if bool(AceFlags.INHERIT_ONLY_ACE & ace.AceFlags) is True:
			continue

		if ace.AceType == ACEType.ACCESS_ALLOWED_ACE_TYPE:
			if group_lookup(ace.Sid, sid_groups) is True:
				if max_allowed_mode is True:
					allowed_access |= ace.Mask
					granted_access |= ace.Mask
				else:
					remainging_access &= ~ace.Mask
					granted_access |= (remainging_access & ace.Mask)

		elif ace.AceType == ACEType.ACCESS_DENIED_ACE_TYPE:
			#print('AM %s '% ace.Mask)
			if ace.Mask == 0:
				break
			if group_lookup(ace.Sid, sid_groups) is True:
				#print('here')
				if max_allowed_mode is True:
					denied_access |= ace.Mask
				else:
					#print(bin(remainging_access))
					#print(bin(ace.Mask))
					if (remainging_access & ace.Mask) != 0:
						granted_access = 0
						return (False, granted_access)
		
	
	if max_allowed_mode is True:
		granted_access = allowed_access & ~denied_access
		if granted_access != 0:
			return True, granted_access
		return False, granted_access
	
	granted_access = 0
	if remainging_access == 0:
		return True, granted_access
	return False, granted_access



from winacl.dtyp.wcee.rsaprivkey import RSAPrivateKeyBlob, BCRYPTRSAKeyBlob
from winacl.dtyp.wcee.ecdhprivkey import ECDHPrivateKeyBlob

# https://learn.microsoft.com/en-us/windows/win32/seccng/key-storage-and-retrieval
# ECDH and ECDSA structures are basically the same, so we can use the same class for both

def keyblobselector(buff):
    magic = buff.read(4)
    buff.seek(-4, 1)
    magic_int = int.from_bytes(magic, byteorder='little', signed=False)
    magic_int_big = int.from_bytes(magic, byteorder='big', signed=False)
    magic_str = None
    try:
        magic_str = magic.decode('ascii')
        magic_str_rev = magic[::-1].decode('ascii')
    except:
        pass
    
    if magic == b'\x07\x02\x00\x00':
        return RSAPrivateKeyBlob
    if magic_str in magic_obj_map:
        return magic_obj_map[magic_str]
    elif magic_str_rev in magic_obj_map:
        return magic_obj_map[magic_str_rev]
    else:
        raise Exception('Unknown key type ASCII: "%s" ASCII_REV: "%s" HEX_LITTLE: %s, HEX_BIG: %s' % (magic_str, magic_str_rev, magic.hex(), magic[::-1].hex()))


# https://github.com/tpn/winsdk-10/blob/master/Include/10.0.14393.0/shared/bcrypt.h
# ints are kept so ppl can find it easier
magics = {
    'ECK1': 0x314B4345, # The key is a 256-bit ECDH public key. BCRYPT_ECDH_PUBLIC_P256_MAGIC
    'ECK2': 0x324B4345, # The key is a 256-bit ECDH private key. BCRYPT_ECDH_PRIVATE_P256_MAGIC
    'ECK3': 0x334B4345, # The key is a 384-bit ECDH public key. BCRYPT_ECDH_PUBLIC_P384_MAGIC
    'ECK4': 0x344B4345, # The key is a 384-bit ECDH private key. BCRYPT_ECDH_PRIVATE_P384_MAGIC
    'ECK5': 0x354B4345, # The key is a 521-bit ECDH public key. BCRYPT_ECDH_PUBLIC_P521_MAGIC
    'ECK6': 0x364B4345, # The key is a 521-bit ECDH private key. BCRYPT_ECDH_PRIVATE_P521_MAGIC
    'ECKP': 0x504B4345, # BCRYPT_ECDH_PUBLIC_GENERIC_MAGIC
    'ECKV': 0x564B4345, # BCRYPT_ECDH_PRIVATE_GENERIC_MAGIC
    'ECS1': 0x31534345, # The key is an ECDSA public key. BCRYPT_ECDSA_PUBLIC_P256_MAGIC
    'ECS2': 0x32534345, # The key is an ECDSA private key. BCRYPT_ECDSA_PRIVATE_P256_MAGIC
    'ECS3': 0x33534345, # The key is an ECDSA public key. BCRYPT_ECDSA_PUBLIC_P384_MAGIC
    'ECS4': 0x34534345, # The key is an ECDSA private key. BCRYPT_ECDSA_PRIVATE_P384_MAGIC
    'ECS5': 0x35534345, # The key is an ECDSA public key. BCRYPT_ECDSA_PUBLIC_P521_MAGIC
    'ECS6': 0x36534345, # The key is an ECDSA private key. BCRYPT_ECDSA_PRIVATE_P521_MAGIC
    'ECDP': 0x50444345, # BCRYPT_ECDSA_PUBLIC_GENERIC_MAGIC
    'ECDV': 0x56444345, # BCRYPT_ECDSA_PRIVATE_GENERIC_MAGIC
    'RSA1': 0x31415352, # The key is an RSA public key. BCRYPT_RSAPUBLIC_MAGIC
    'RSA2': 0x32415352, # The key is an RSA private key. BCRYPT_RSAPRIVATE_MAGIC
    'RSA3': 0x33415352, # The key is an RSA key pair with parameters. BCRYPT_RSAFULLPRIVATE_MAGIC
}

magic_obj_map = {
    'ECK1': ECDHPrivateKeyBlob, # The key is a 256-bit ECDH public key. BCRYPT_ECDH_PUBLIC_P256_MAGIC
    'ECK2': ECDHPrivateKeyBlob,
    'ECK3': ECDHPrivateKeyBlob,
    'ECK4': ECDHPrivateKeyBlob,
    'ECK5': ECDHPrivateKeyBlob,
    'ECK6': ECDHPrivateKeyBlob,
    'ECKP': ECDHPrivateKeyBlob,
    'ECKV': ECDHPrivateKeyBlob,
    'ECS1': ECDHPrivateKeyBlob,
    'ECS2': ECDHPrivateKeyBlob,
    'ECS3': ECDHPrivateKeyBlob,
    'ECS4': ECDHPrivateKeyBlob,
    'ECS5': ECDHPrivateKeyBlob,
    'ECS6': ECDHPrivateKeyBlob,
    'ECDP': ECDHPrivateKeyBlob,
    'ECDV': ECDHPrivateKeyBlob,
    'RSA1': BCRYPTRSAKeyBlob,
    'RSA2': BCRYPTRSAKeyBlob,
    'RSA3': BCRYPTRSAKeyBlob,
}


import enum

class SE_OBJECT_TYPE(enum.Enum):
	SE_UNKNOWN_OBJECT_TYPE = 0 #Unknown object type.
	SE_FILE_OBJECT = 1 #Indicates a file or directory.
	SE_SERVICE = 2 #Indicates a Windows service
	SE_PRINTER = 3 #Indicates a printer.
	SE_REGISTRY_KEY = 4 #Indicates a registry key.
	SE_LMSHARE = 5 #Indicates a network share.
	SE_KERNEL_OBJECT = 6 #Indicates a local 
	SE_WINDOW_OBJECT = 7 #Indicates a window station or desktop object on the local computer
	SE_DS_OBJECT = 8 #Indicates a directory service object or a property set or property of a directory service object. 
	SE_DS_OBJECT_ALL = 9 #Indicates a directory service object and all of its property sets and properties.
	SE_PROVIDER_DEFINED_OBJECT = 10 #Indicates a provider-defined object.
	SE_WMIGUID_OBJECT = 11 #Indicates a WMI object.
	SE_REGISTRY_WOW64_32KEY = 12 #Indicates an object for a registry entry under WOW64.
	SE_REGISTRY_WOW64_64KEY = 13 #Indicates an object for a registry entry under WOW64.


import io

class PREFERRED_BACKUP_KEY:
	def __init__(self):
		self.version = None
		self.keylength = None
		self.certificatelength = None
		self.keydata = None
		self.certdata = None

	@staticmethod
	def from_bytes(data):
		return PREFERRED_BACKUP_KEY.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		pbk = PREFERRED_BACKUP_KEY()
		pbk.version = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		pbk.keylength = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		pbk.certificatelength = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		pbk.keydata = buff.read(pbk.keylength)
		pbk.certdata = buff.read(pbk.certificatelength)
		return pbk


import enum
from winacl.functions.defines import *
from winacl.functions.kernel32 import GetLastError, LocalFree, READ_CONTROL
from winacl.functions.membuff import MemoryBuffer
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR
from winacl.dtyp.sid import SID
from winacl.functions.constants import SE_OBJECT_TYPE

def win_succ_check(result, func, arguments):
	if result != 0:
		raise ctypes.WinError(result)
	return result

### https://github.com/MarioVilas/winappdbg/blob/master/winappdbg/win32/advapi32.py
### most parts taken from here!

TOKEN_INFORMATION_CLASS = ctypes.c_int

TokenUser                               = 1
TokenGroups                             = 2
TokenPrivileges                         = 3
TokenOwner                              = 4
TokenPrimaryGroup                       = 5
TokenDefaultDacl                        = 6
TokenSource                             = 7
TokenType                               = 8
TokenImpersonationLevel                 = 9
TokenStatistics                         = 10
TokenRestrictedSids                     = 11
TokenSessionId                          = 12
TokenGroupsAndPrivileges                = 13
TokenSessionReference                   = 14
TokenSandBoxInert                       = 15
TokenAuditPolicy                        = 16
TokenOrigin                             = 17
TokenElevationType                      = 18
TokenLinkedToken                        = 19
TokenElevation                          = 20
TokenHasRestrictions                    = 21
TokenAccessInformation                  = 22
TokenVirtualizationAllowed              = 23
TokenVirtualizationEnabled              = 24
TokenIntegrityLevel                     = 25
TokenUIAccess                           = 26
TokenMandatoryPolicy                    = 27
TokenLogonSid                           = 28
TokenIsAppContainer                     = 29
TokenCapabilities                       = 30
TokenAppContainerSid                    = 31
TokenAppContainerNumber                 = 32
TokenUserClaimAttributes                = 33
TokenDeviceClaimAttributes              = 34
TokenRestrictedUserClaimAttributes      = 35
TokenRestrictedDeviceClaimAttributes    = 36
TokenDeviceGroups                       = 37
TokenRestrictedDeviceGroups             = 38
TokenSecurityAttributes                 = 39
TokenIsRestricted                       = 40
MaxTokenInfoClass                       = 41

# Token access rights
TOKEN_ASSIGN_PRIMARY    = 0x0001
TOKEN_DUPLICATE         = 0x0002
TOKEN_IMPERSONATE       = 0x0004
TOKEN_QUERY             = 0x0008
TOKEN_QUERY_SOURCE      = 0x0010
TOKEN_ADJUST_PRIVILEGES = 0x0020
TOKEN_ADJUST_GROUPS     = 0x0040
TOKEN_ADJUST_DEFAULT    = 0x0080
TOKEN_ADJUST_SESSIONID  = 0x0100
TOKEN_READ = (STANDARD_RIGHTS_READ | TOKEN_QUERY)
TOKEN_ALL_ACCESS = (STANDARD_RIGHTS_REQUIRED | TOKEN_ASSIGN_PRIMARY |
		TOKEN_DUPLICATE | TOKEN_IMPERSONATE | TOKEN_QUERY | TOKEN_QUERY_SOURCE |
		TOKEN_ADJUST_PRIVILEGES | TOKEN_ADJUST_GROUPS | TOKEN_ADJUST_DEFAULT |
		TOKEN_ADJUST_SESSIONID)

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/23e75ca3-98fd-4396-84e5-86cd9d40d343
OWNER_SECURITY_INFORMATION = 0x00000001 #The owner identifier of the object is being referenced.
GROUP_SECURITY_INFORMATION = 0x00000002 #The primary group identifier of the object is being referenced.
DACL_SECURITY_INFORMATION = 0x00000004 #The DACL of the object is being referenced.
SACL_SECURITY_INFORMATION = 0x00000008 #The SACL of the object is being referenced.
LABEL_SECURITY_INFORMATION = 0x00000010 #The mandatory integrity label is being referenced.
UNPROTECTED_SACL_SECURITY_INFORMATION = 0x10000000 #The SACL inherits access control entries (ACEs) from the parent object.
UNPROTECTED_DACL_SECURITY_INFORMATION = 0x20000000 #The DACL inherits ACEs from the parent object.
PROTECTED_SACL_SECURITY_INFORMATION = 0x40000000 #The SACL cannot inherit ACEs.
PROTECTED_DACL_SECURITY_INFORMATION = 0x80000000 #The DACL cannot inherit ACEs.
ATTRIBUTE_SECURITY_INFORMATION = 0x00000020 #A SYSTEM_RESOURCE_ATTRIBUTE_ACE (section 2.4.4.15) is being referenced.
SCOPE_SECURITY_INFORMATION = 0x00000040 #A SYSTEM_SCOPED_POLICY_ID_ACE (section 2.4.4.16) is being referenced.
PROCESS_TRUST_LABEL_SECURITY_INFORMATION = 0x00000080 #Reserved.
BACKUP_SECURITY_INFORMATION = 0x00010000 #The security descriptor is being accessed for use in a backup operation.


# EnumServicesStatusEx() service type filters (in addition to actual types)
SERVICE_DRIVER = 0x0000000B # SERVICE_KERNEL_DRIVER and SERVICE_FILE_SYSTEM_DRIVER
SERVICE_WIN32  = 0x00000030 # SERVICE_WIN32_OWN_PROCESS and SERVICE_WIN32_SHARE_PROCESS

# EnumServicesStatusEx() service state filters
SERVICE_ACTIVE    = 1
SERVICE_INACTIVE  = 2
SERVICE_STATE_ALL = 3


SC_HANDLE = HANDLE

SERVICE_ALL_ACCESS           = 0xF01FF
SERVICE_QUERY_CONFIG         = 0x0001
SERVICE_CHANGE_CONFIG        = 0x0002
SERVICE_QUERY_STATUS         = 0x0004
SERVICE_ENUMERATE_DEPENDENTS = 0x0008
SERVICE_START                = 0x0010
SERVICE_STOP                 = 0x0020
SERVICE_PAUSE_CONTINUE       = 0x0040
SERVICE_INTERROGATE          = 0x0080
SERVICE_USER_DEFINED_CONTROL = 0x0100

SC_MANAGER_ALL_ACCESS           = 0xF003F
SC_MANAGER_CONNECT              = 0x0001
SC_MANAGER_CREATE_SERVICE       = 0x0002
SC_MANAGER_ENUMERATE_SERVICE    = 0x0004
SC_MANAGER_LOCK                 = 0x0008
SC_MANAGER_QUERY_LOCK_STATUS    = 0x0010
SC_MANAGER_MODIFY_BOOT_CONFIG   = 0x0020



# https://docs.microsoft.com/en-us/windows/win32/api/aclapi/nf-aclapi-setsecurityinfo

#DWORD SetSecurityInfo(
#  HANDLE               handle,
#  SE_OBJECT_TYPE       ObjectType,
#  SECURITY_INFORMATION SecurityInfo,
#  PSID                 psidOwner,
#  PSID                 psidGroup,
#  PACL                 pDacl,
#  PACL                 pSacl
#);


def SetSecurityInfo(handle_obj, obj_type, sd):
	#Warning  If the supplied handle was opened with an ACCESS_MASK value of MAXIMUM_ALLOWED, then the SetSecurityInfo function will not propagate ACEs to children.

	_SetSecurityInfo = windll.advapi32.SetSecurityInfo
	_SetSecurityInfo.argtypes = [HANDLE, DWORD, DWORD, PVOID, PVOID, PVOID, PVOID]
	_SetSecurityInfo.restype  = DWORD
	_SetSecurityInfo.errcheck = RaiseIfNotErrorSuccess

	sec_info = 0
	pOwner = None
	if sd.Owner is not None:
		sec_info |= OWNER_SECURITY_INFORMATION
		owner_data = sd.Owner.to_bytes()
		pOwner = ctypes.create_string_buffer(owner_data, len(owner_data))
	
	pGroup = None
	#if sd.Group is not None:
	#	sec_info |= GROUP_SECURITY_INFORMATION
	#	group_data = sd.Group.to_bytes()
	#	pGroup = ctypes.create_string_buffer(group_data, len(group_data))
	
	pDacl = None
	if sd.Dacl is not None:
		sec_info |= DACL_SECURITY_INFORMATION
		dacl_data = sd.Dacl.to_bytes()
		pDacl = ctypes.create_string_buffer(dacl_data, len(dacl_data))
	
	pSacl = None
	#if sd.Sacl is not None:
	#	sec_info |= SACL_SECURITY_INFORMATION
	#	sacl_data = sd.Sacl.to_bytes()
	#	pSacl = ctypes.create_string_buffer(sacl_data, len(sacl_data))

	if sec_info == 0:
		raise Exception('Looks like nothing is to be set! Check your sd object!')

	#print(sec_info)
	ret = _SetSecurityInfo(
		handle_obj, 
		obj_type, 
		sec_info, 
		pOwner, 
		pGroup, 
		pDacl, 
		pSacl
	)




# https://docs.microsoft.com/en-us/windows/win32/api/aclapi/nf-aclapi-getsecurityinfo

# DWORD GetSecurityInfo(
#  HANDLE               handle,
#  SE_OBJECT_TYPE       ObjectType,
#  SECURITY_INFORMATION SecurityInfo,
#  PSID                 *ppsidOwner,
#  PSID                 *ppsidGroup,
#  PACL                 *ppDacl,
#  PACL                 *ppSacl,
#  PSECURITY_DESCRIPTOR *ppSecurityDescriptor
#);

def GetSecurityInfo(handle_obj, obj_type, info_req):
	_GetSecurityInfo = windll.advapi32.GetSecurityInfo
	_GetSecurityInfo.argtypes = [HANDLE, DWORD, DWORD, PVOID, PVOID, PVOID, PVOID, PVOID] #[HANDLE, SE_OBJECT_TYPE, DWORD, PSID, PSID, PACL, PACL, PSECURITY_DESCRIPTOR]
	_GetSecurityInfo.restype  = DWORD
	_GetSecurityInfo.errcheck = win_succ_check

	ppsidOwner = None
	ppsidGroup = None
	ppDacl = None
	ppSacl = None
	ppSecurityDescriptor = ctypes.pointer(ctypes.c_uint(0))

	ret = _GetSecurityInfo(handle_obj, obj_type, info_req, ppsidOwner, ppsidGroup, ppDacl, ppSacl, ctypes.byref(ppSecurityDescriptor))
	pSecurityDescriptor = ppSecurityDescriptor.contents
	buff = MemoryBuffer(ctypes.addressof(ppSecurityDescriptor.contents))	
	sd = SECURITY_DESCRIPTOR.from_buffer(buff, obj_type)
	LocalFree(ppSecurityDescriptor)
	return sd


# https://docs.microsoft.com/en-us/windows/win32/api/accctrl/ns-accctrl-trustee_w
class TRUSTEE_W(Structure):
	_fields_ = [
		("pMultipleTrustee", PVOID),
		("MultipleTrusteeOperation", DWORD),
		("TrusteeForm", DWORD),
		("TrusteeType", DWORD),
		("ptstrName", HANDLE),
	]
PTRUSTEE_W = ctypes.POINTER(TRUSTEE_W)


# https://docs.microsoft.com/en-us/windows/win32/api/aclapi/nf-aclapi-buildtrusteewithsidw
#void BuildTrusteeWithSidW(
#  PTRUSTEE_W pTrustee,
#  PSID       pSid
#);

def BuildTrusteeWithSidW(sid):
	_BuildTrusteeWithSidW = windll.advapi32.BuildTrusteeWithSidW
	_BuildTrusteeWithSidW.argtypes = [PTRUSTEE_W, PVOID] #[HANDLE, SE_OBJECT_TYPE, DWORD, PSID, PSID, PACL, PACL, PSECURITY_DESCRIPTOR]
	_BuildTrusteeWithSidW.restype  = None

	sid_data = sid.to_bytes()
	csid = ctypes.create_string_buffer(sid_data, len(sid_data))

	trustee = TRUSTEE_W()

	_BuildTrusteeWithSidW(trustee, csid)
	return trustee

# https://docs.microsoft.com/en-us/windows/win32/api/sddl/nf-sddl-convertsidtostringsidw

#BOOL ConvertSidToStringSidW(
#  PSID   Sid,
#  LPWSTR *StringSid
#);

def ConvertSidToStringSidW(sid, is_raw_ptr = False):
	_ConvertSidToStringSidW = windll.advapi32.ConvertSidToStringSidW
	_ConvertSidToStringSidW.argtypes = [PVOID, PVOID] #[HANDLE, SE_OBJECT_TYPE, DWORD, PSID, PSID, PACL, PACL, PSECURITY_DESCRIPTOR]
	_ConvertSidToStringSidW.restype  = DWORD
	_ConvertSidToStringSidW.errcheck = RaiseIfZero

	if is_raw_ptr is False:
		sid_data = sid.to_bytes()
		csid = ctypes.create_string_buffer(sid_data, len(sid_data))
		pstr = ctypes.create_unicode_buffer(1) #size is irrelevant here
		cstr_sid = ctypes.pointer(pstr)
	
		_ConvertSidToStringSidW(byref(csid), byref(cstr_sid))
	else:
		pstr = ctypes.create_unicode_buffer(1) #size is irrelevant here
		cstr_sid = ctypes.pointer(pstr)
	
		_ConvertSidToStringSidW(sid, byref(cstr_sid))

	str_sid = ctypes.wstring_at(cstr_sid)
	LocalFree(cstr_sid)

	return str_sid

# https://docs.microsoft.com/en-us/windows/win32/api/sddl/nf-sddl-convertstringsidtosidw

#BOOL ConvertStringSidToSidW(
#  LPCWSTR StringSid,
#  PSID   *Sid
#);
def ConvertStringSidToSidW(sid_str):
	_ConvertStringSidToSidW = windll.advapi32.ConvertStringSidToSidW
	_ConvertStringSidToSidW.argtypes = [PVOID, PVOID] #[HANDLE, SE_OBJECT_TYPE, DWORD, PSID, PSID, PACL, PACL, PSECURITY_DESCRIPTOR]
	_ConvertStringSidToSidW.restype  = DWORD
	_ConvertStringSidToSidW.errcheck = RaiseIfZero

	

	cstr_sid = ctypes.create_string_buffer(sid_str.encode('utf-16-le'))
	ppSecurityDescriptor = ctypes.pointer(ctypes.c_uint(0))
	
	_ConvertStringSidToSidW(cstr_sid, byref(ppSecurityDescriptor))
	buff = MemoryBuffer(ctypes.addressof(ppSecurityDescriptor.contents))
	sd = SID.from_buffer(buff)
	LocalFree(ppSecurityDescriptor)
	return sd

# https://docs.microsoft.com/en-us/windows/win32/api/aclapi/nf-aclapi-geteffectiverightsfromaclw
#DWORD GetEffectiveRightsFromAclW(
#  PACL         pacl,
#  PTRUSTEE_W   pTrustee,
#  PACCESS_MASK pAccessRights
#);

def GetEffectiveRightsFromAclW(acl, sid):
	"""
	Takes a SID instead of a trustee!
	"""
	_GetEffectiveRightsFromAclW = windll.advapi32.GetEffectiveRightsFromAclW
	_GetEffectiveRightsFromAclW.argtypes = [PVOID, PTRUSTEE_W, PDWORD] #[HANDLE, SE_OBJECT_TYPE, DWORD, PSID, PSID, PACL, PACL, PSECURITY_DESCRIPTOR]
	_GetEffectiveRightsFromAclW.restype  = RaiseIfNotErrorSuccess

	sid_data = sid.to_bytes()
	psid = ctypes.create_string_buffer(sid_data, len(sid_data))
	trustee = TRUSTEE_W()
	trustee.pMultipleTrustee = 0
	trustee.MultipleTrusteeOperation = 0
	trustee.TrusteeForm = 0
	trustee.TrusteeType = 0
	trustee.ptstrName = ctypes.c_void_p(ctypes.addressof(psid))

	effective_rigths_mask = DWORD(0)
	acl_data = acl.to_bytes()
	pacl = ctypes.create_string_buffer(acl_data, len(acl_data))

	res = _GetEffectiveRightsFromAclW(pacl, trustee, byref(effective_rigths_mask))
	return effective_rigths_mask.value

def LookupAccountSidW(lpSystemName, sid_data):
	_LookupAccountSidW = windll.advapi32.LookupAccountSidW
	_LookupAccountSidW.argtypes = [LPSTR, PSID, LPWSTR, LPDWORD, LPWSTR, LPDWORD, LPDWORD]
	_LookupAccountSidW.restype  = bool

	cchName = DWORD(0)
	cchReferencedDomainName = DWORD(0)
	peUse = DWORD(0)
	lpSid = ctypes.create_string_buffer(sid_data, len(sid_data))
	_LookupAccountSidW(lpSystemName, lpSid, None, byref(cchName), None, byref(cchReferencedDomainName), byref(peUse))
	error = GetLastError()
	if error != ERROR_INSUFFICIENT_BUFFER:
		raise ctypes.WinError(error)
	lpName = ctypes.create_unicode_buffer(u'', cchName.value + 1)
	lpReferencedDomainName = ctypes.create_unicode_buffer(u'', cchReferencedDomainName.value + 1)
	success = _LookupAccountSidW(lpSystemName, lpSid, lpName, byref(cchName), lpReferencedDomainName, byref(cchReferencedDomainName), byref(peUse))
	if not success:
		raise ctypes.WinError()
	return lpName.value, lpReferencedDomainName.value, peUse.value

# https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-lookupaccountnamew
def LookupAccountNameW(lpSystemName, accountname):
	_LookupAccountNameW = windll.advapi32.LookupAccountNameW
	_LookupAccountNameW.argtypes = [LPWSTR, LPWSTR, PSID, LPDWORD, LPWSTR, LPDWORD, LPDWORD]
	_LookupAccountNameW.restype  = BOOL

	cbSid = DWORD(0)
	cchReferencedDomainName = DWORD(0)
	peUse = DWORD(0)
	lpAccountName = ctypes.create_unicode_buffer(accountname)
	_LookupAccountNameW(lpSystemName, lpAccountName, None, byref(cbSid), None, byref(cchReferencedDomainName), byref(peUse))
	error = GetLastError()
	if error != ERROR_INSUFFICIENT_BUFFER:
		raise(ctypes.WinError(error))
	sid = ctypes.create_string_buffer(b'', cbSid.value)
	psid = ctypes.cast(ctypes.pointer(sid), PSID)
	lpReferencedDomainName = ctypes.create_unicode_buffer(u'', cchReferencedDomainName.value + 1)
	success = _LookupAccountNameW(lpSystemName, lpAccountName, psid, byref(cbSid), lpReferencedDomainName, byref(cchReferencedDomainName), byref(peUse))
	if not success:
		raise ctypes.WinError()
	
	buff = MemoryBuffer(psid.value)
	sid = SID.from_buffer(buff)
	#LocalFree(psid)
	return sid, lpReferencedDomainName.value, peUse.value


# https://docs.microsoft.com/en-us/windows/win32/api/sddl/nf-sddl-convertsecuritydescriptortostringsecuritydescriptorw
# https://docs.microsoft.com/en-us/windows/win32/secauthz/ace-strings
# https://docs.microsoft.com/en-us/windows/win32/secauthz/security-descriptor-string-format
SDDL_REVISION_1 = 1

def ConvertSecurityDescriptorToStringSecurityDescriptorW(sd, req_info = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION):
	_ConvertSecurityDescriptorToStringSecurityDescriptorW = windll.advapi32.ConvertSecurityDescriptorToStringSecurityDescriptorW
	_ConvertSecurityDescriptorToStringSecurityDescriptorW.argtypes = [PVOID, DWORD, DWORD, PVOID, PULONG]
	_ConvertSecurityDescriptorToStringSecurityDescriptorW.restype  = DWORD
	_ConvertSecurityDescriptorToStringSecurityDescriptorW.errcheck  = RaiseIfZero

	sd_data = sd.to_bytes()
	psd = ctypes.create_string_buffer(sd_data, len(sd_data))
	pstr = ctypes.pointer(ctypes.c_uint(0))
	pstrsize = ctypes.c_uint(0)

	_ConvertSecurityDescriptorToStringSecurityDescriptorW(psd, SDDL_REVISION_1, req_info, byref(pstr), byref(pstrsize))

	res = ctypes.string_at(pstr, pstrsize.value*2) #multiply by two because the size is not in bytes but in character counts and each char is 2 bytes.... fuck you microsoft
	LocalFree(pstr)
	return res.decode('utf-16-le')

#BOOL QueryServiceObjectSecurity(
#  SC_HANDLE            hService,
#  SECURITY_INFORMATION dwSecurityInformation,
#  PSECURITY_DESCRIPTOR lpSecurityDescriptor,
#  DWORD                cbBufSize,
#  LPDWORD              pcbBytesNeeded
#);

def QueryServiceObjectSecurity(hService, dwSecurityInformation = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION, sd_object_type = None):
	_QueryServiceObjectSecurity = windll.advapi32.QueryServiceObjectSecurity
	_QueryServiceObjectSecurity.argtypes = [SC_HANDLE, DWORD, PVOID, DWORD, LPDWORD]
	_QueryServiceObjectSecurity.restype  = DWORD
	#_QueryServiceObjectSecurity.errcheck = RaiseIfZero

	#first we get the size
	lpSecurityDescriptor = None
	cbBufSize = DWORD(0)
	pcbBytesNeeded = DWORD(0)

	correct_length = 0
	res = _QueryServiceObjectSecurity(hService, dwSecurityInformation, lpSecurityDescriptor, cbBufSize, pcbBytesNeeded)
	if res == 0:
		#getting the correct length
		correct_length = pcbBytesNeeded.value
	
	lpSecurityDescriptor = ctypes.create_string_buffer(correct_length)
	cbBufSize = DWORD(correct_length)
	pcbBytesNeeded = DWORD(0)

	res = _QueryServiceObjectSecurity(hService, dwSecurityInformation, lpSecurityDescriptor, cbBufSize, pcbBytesNeeded)
	if res == 0:
		raise ctypes.WinError(result)
	buff = ctypes.string_at(lpSecurityDescriptor, pcbBytesNeeded.value)
	sd = SECURITY_DESCRIPTOR.from_bytes(buff, sd_object_type)

	return sd


def OpenSCManagerW(lpMachineName = None, lpDatabaseName = None, dwDesiredAccess = SC_MANAGER_ALL_ACCESS):
	_OpenSCManagerW = windll.advapi32.OpenSCManagerW
	_OpenSCManagerW.argtypes = [LPWSTR, LPWSTR, DWORD]
	_OpenSCManagerW.restype  = SC_HANDLE
	_OpenSCManagerW.errcheck = RaiseIfZero

	hSCObject = _OpenSCManagerW(lpMachineName, lpDatabaseName, dwDesiredAccess)
	return hSCObject

#
#SC_HANDLE OpenServiceW(
#  SC_HANDLE hSCManager,
#  LPCSTR    lpServiceName,
#  DWORD     dwDesiredAccess
#);

def OpenServiceW(hSCManager, lpServiceName, dwDesiredAccess = SERVICE_ALL_ACCESS):
	_OpenServiceW = windll.advapi32.OpenServiceW
	_OpenServiceW.argtypes = [SC_HANDLE, LPWSTR, DWORD]
	_OpenServiceW.restype  = SC_HANDLE
	_OpenServiceW.errcheck = RaiseIfZero
	return _OpenServiceW(hSCManager, lpServiceName, dwDesiredAccess)

def CloseServiceHandle(hHandle):
	_CloseServiceHandle = windll.advapi32.CloseServiceHandle
	_CloseServiceHandle.argtypes = [HANDLE]
	_CloseServiceHandle.restype  = bool
	_CloseServiceHandle.errcheck = RaiseIfZero
	_CloseServiceHandle(hHandle)


class SERVICE_STATUS(Structure):
	_fields_ = [
		("dwServiceType",               DWORD),
		("dwCurrentState",              DWORD),
		("dwControlsAccepted",          DWORD),
		("dwWin32ExitCode",             DWORD),
		("dwServiceSpecificExitCode",   DWORD),
		("dwCheckPoint",                DWORD),
		("dwWaitHint",                  DWORD),
	]
LPSERVICE_STATUS = POINTER(SERVICE_STATUS)

class ENUM_SERVICE_STATUSW(Structure):
	_fields_ = [
		("lpServiceName", LPWSTR),
		("lpDisplayName", LPWSTR),
		("ServiceStatus", SERVICE_STATUS),
	]
LPENUM_SERVICE_STATUSW = POINTER(ENUM_SERVICE_STATUSW)

def EnumServicesStatusW(hSCManager, dwServiceType = SERVICE_DRIVER | SERVICE_WIN32, dwServiceState = SERVICE_STATE_ALL):
	_EnumServicesStatusW = windll.advapi32.EnumServicesStatusW
	_EnumServicesStatusW.argtypes = [SC_HANDLE, DWORD, DWORD, LPVOID, DWORD, LPDWORD, LPDWORD, LPDWORD]
	_EnumServicesStatusW.restype  = bool

	cbBytesNeeded    = DWORD(0)
	ServicesReturned = DWORD(0)
	ResumeHandle     = DWORD(0)

	_EnumServicesStatusW(hSCManager, dwServiceType, dwServiceState, None, 0, byref(cbBytesNeeded), byref(ServicesReturned), byref(ResumeHandle))

	Services = []
	success = False
	while GetLastError() == ERROR_MORE_DATA:
		if cbBytesNeeded.value < sizeof(ENUM_SERVICE_STATUSW):
			break
		ServicesBuffer = ctypes.create_string_buffer(b"", cbBytesNeeded.value)
		success = _EnumServicesStatusW(hSCManager, dwServiceType, dwServiceState, byref(ServicesBuffer), sizeof(ServicesBuffer), byref(cbBytesNeeded), byref(ServicesReturned), byref(ResumeHandle))
		if sizeof(ServicesBuffer) < (sizeof(ENUM_SERVICE_STATUSW) * ServicesReturned.value):
			raise ctypes.WinError()
		lpServicesArray = ctypes.cast(ctypes.cast(ctypes.pointer(ServicesBuffer), ctypes.c_void_p), LPENUM_SERVICE_STATUSW)
		for index in range(0, ServicesReturned.value):
			Services.append( lpServicesArray[index].lpServiceName )
		if success: break
	if not success:
		raise ctypes.WinError()

	return Services


HKEY_CLASSES_ROOT = 0x80000000
HKEY_CURRENT_USER = 0x80000001
HKEY_LOCAL_MACHINE = 0x80000002
HKEY_USERS = 0x80000003
HKEY_PERFORMANCE_DATA = 0x80000004
HKEY_PERFORMANCE_TEXT = 0x80000050
HKEY_PERFORMANCE_NLSTEXT = 0x80000060
HKEY_CURRENT_CONFIG = 0x80000005
HKEY_DYN_DATA = 0x80000006

hive_name_map = {
	'HKLM' : HKEY_LOCAL_MACHINE,
	'HKCU' : HKEY_CURRENT_USER,
	'HKCR' : HKEY_CLASSES_ROOT,
	'HKU'  : HKEY_USERS,
	'HKCC' : HKEY_CURRENT_CONFIG,
	'HKPD' : HKEY_PERFORMANCE_DATA,
}

# https://docs.microsoft.com/en-us/windows/win32/sysinfo/predefined-keys
KEY_ALL_ACCESS = 0xF003F #Combines the STANDARD_RIGHTS_REQUIRED, KEY_QUERY_VALUE, KEY_SET_VALUE, KEY_CREATE_SUB_KEY, KEY_ENUMERATE_SUB_KEYS, KEY_NOTIFY, and KEY_CREATE_LINK access rights.
KEY_CREATE_LINK = 0x0020 #	Reserved for system use.
KEY_CREATE_SUB_KEY = 0x0004 #Required to create a subkey of a registry key.
KEY_ENUMERATE_SUB_KEYS = 0x0008 #Required to enumerate the subkeys of a registry key.
KEY_EXECUTE = 0x20019 #Equivalent to KEY_READ.
KEY_NOTIFY = 0x0010 #Required to request change notifications for a registry key or for subkeys of a registry key.
KEY_QUERY_VALUE = 0x0001 #Required to query the values of a registry key.
KEY_READ = 0x20019 #Combines the STANDARD_RIGHTS_READ, KEY_QUERY_VALUE, KEY_ENUMERATE_SUB_KEYS, and KEY_NOTIFY values.
KEY_SET_VALUE = 0x0002 #Required to create, delete, or set a registry value.
KEY_WOW64_32KEY = 0x0200 #Indicates that an application on 64-bit Windows should operate on the 32-bit registry view. This flag is ignored by 32-bit Windows. For more information, see Accessing an Alternate Registry View. #This flag must be combined using the OR operator with the other flags in this table that either query or access registry values. #Windows 2000: This flag is not supported.
KEY_WOW64_64KEY = 0x0100 #Indicates that an application on 64-bit Windows should operate on the 64-bit registry view. This flag is ignored by 32-bit Windows. For more information, see Accessing an Alternate Registry View. This flag must be combined using the OR operator with the other flags in this table that either query or access registry values. Windows 2000: This flag is not supported.
KEY_WRITE = 0x20006 #Combines the STANDARD_RIGHTS_WRITE, KEY_SET_VALUE, and KEY_CREATE_SUB_KEY access rights.


REG_OPTION_RESERVED = 0x0000
REG_OPTION_NON_VOLATILE = 0x0000
REG_OPTION_VOLATILE = 0x0001
REG_OPTION_CREATE_LINK = 0x0002
REG_OPTION_BACKUP_RESTORE = 0x0004
REG_OPTION_OPEN_LINK = 0x0008
#REG_LEGAL_OPTION =
#
#    REG_OPTION_RESERVED |
#    REG_OPTION_NON_VOLATILE | REG_OPTION_CREATE_LINK |
#    REG_OPTION_BACKUP_RESTORE | REG_OPTION_OPEN_LINK

# https://docs.microsoft.com/en-us/windows/win32/api/winreg/nf-winreg-regopenkeyexw
def RegOpenKeyExW(regkey_handle, subkey_name = None, options = REG_OPTION_NON_VOLATILE, access = KEY_READ):
	_RegOpenKeyExW = windll.advapi32.RegOpenKeyExA
	_RegOpenKeyExW.argtypes = [HANDLE, LPVOID, DWORD, DWORD, PHANDLE]
	_RegOpenKeyExW.restype  = DWORD
	_RegOpenKeyExW.errcheck = RaiseIfNotErrorSuccess

	psubkey = None
	if subkey_name is not None:
		psd = subkey_name.encode('ascii') #despite the W at the end of the function name, it takes ASCII
		psubkey = ctypes.create_string_buffer(psd, len(psd))

	reshandle = HANDLE()
	ret = _RegOpenKeyExW(regkey_handle, psubkey, options, access, ctypes.byref(reshandle))
	return reshandle

# https://docs.microsoft.com/en-us/windows/win32/api/winreg/nf-winreg-regclosekey
def RegCloseKey(hHandle):
	_RegCloseKey = windll.advapi32.RegCloseKey
	_RegCloseKey.argtypes = [HANDLE]
	_RegCloseKey.restype  = bool
	_RegCloseKey.errcheck = RaiseIfNotErrorSuccess
	_RegCloseKey(hHandle)


# https://docs.microsoft.com/en-us/windows/win32/api/winreg/nf-winreg-regenumkeyexw
# LSTATUS RegEnumKeyExW(
#  HKEY      hKey,
#  DWORD     dwIndex,
#  LPWSTR    lpName,
#  LPDWORD   lpcchName,
#  LPDWORD   lpReserved,
#  LPWSTR    lpClass,
#  LPDWORD   lpcchClass,
#  PFILETIME lpftLastWriteTime
#);

def RegEnumKeyExW(hKey):
	_RegEnumKeyExW = windll.advapi32.RegEnumKeyExW
	_RegEnumKeyExW.argtypes = [HANDLE, DWORD, PVOID, LPDWORD, LPDWORD, LPWSTR, LPDWORD, PVOID]
	_RegEnumKeyExW.restype  = DWORD

	i = 0
	subkeys = []
	res = ERROR_SUCCESS
	
	while res == ERROR_SUCCESS:
		dwIndex = DWORD(i)
		lpName = ctypes.create_unicode_buffer(255 * 2)
		lpcchName = DWORD(255)

		res = _RegEnumKeyExW(hKey, dwIndex, byref(lpName), byref(lpcchName), NULL, NULL, NULL, NULL)
		if res == ERROR_SUCCESS:
			skname = lpName.value
			subkeys.append(skname)
			i += 1
		elif res == ERROR_NO_MORE_ITEMS:
			break
		else:
			raise ctypes.WinError()

	return subkeys


# https://docs.microsoft.com/en-us/windows/win32/api/winreg/nf-winreg-regenumvaluew
# LSTATUS RegEnumValueW(
#  HKEY    hKey,
#  DWORD   dwIndex,
#  LPWSTR  lpValueName,
#  LPDWORD lpcchValueName,
#  LPDWORD lpReserved,
#  LPDWORD lpType,
#  LPBYTE  lpData,
#  LPDWORD lpcbData
#);

def RegEnumValueW(hKey):
	_RegEnumValueW = windll.advapi32.RegEnumValueW
	_RegEnumValueW.argtypes = [HANDLE, DWORD, PVOID, LPDWORD, PVOID, PVOID, PVOID, PVOID]
	_RegEnumValueW.restype  = DWORD

	i = 0
	subkeys = []
	res = ERROR_SUCCESS
	
	while res == ERROR_SUCCESS:
		dwIndex = DWORD(i)
		lpName = ctypes.create_unicode_buffer(255 * 2)
		lpcchName = DWORD(255)

		res = _RegEnumValueW(hKey, dwIndex, byref(lpName), byref(lpcchName), NULL, NULL, NULL, NULL)
		if res == ERROR_SUCCESS:
			skname = lpName.value
			subkeys.append(skname)
			i += 1
		elif res == ERROR_NO_MORE_ITEMS:
			break
		else:
			raise ctypes.WinError()

	return subkeys


# https://docs.microsoft.com/en-us/windows/win32/api/lmshare/nf-lmshare-netsharegetinfo
#NET_API_STATUS NET_API_FUNCTION NetShareGetInfo(
#  LMSTR  servername,
#  LMSTR  netname,
#  DWORD  level,
#  LPBYTE *bufptr
#);

def NetShareGetInfo(servername, sharename, level = 502):
	_NetShareGetInfo = windll.netapi32.NetShareGetInfo
	_NetShareGetInfo.argtypes = [PVOID, PVOID, DWORD, PVOID]
	_NetShareGetInfo.restype  = DWORD
	_NetShareGetInfo.errcheck = RaiseIfNotErrorSuccess

	pservername = ctypes.create_string_buffer(servername.encode('ascii'), len(servername.encode('ascii')))
	psharename = ctypes.create_string_buffer(sharename.encode('ascii'), len(sharename.encode('ascii')))
	pbuff = ctypes.pointer(ctypes.c_uint(0))

	_NetShareGetInfo(ctypes.pointer((pservername)), ctypes.pointer(psharename), level, pbuff)
	print(pbuff)
	#LocalFree(cstr_sid)

	return str_sid


# BOOL WINAPI OpenProcessToken(
#   __in   HANDLE ProcessHandle,
#   __in   DWORD DesiredAccess,
#   __out  PHANDLE TokenHandle
# );
def OpenProcessToken(ProcessHandle, DesiredAccess = TOKEN_ALL_ACCESS):
	_OpenProcessToken = windll.advapi32.OpenProcessToken
	_OpenProcessToken.argtypes = [HANDLE, DWORD, PHANDLE]
	_OpenProcessToken.restype  = bool
	_OpenProcessToken.errcheck = RaiseIfZero

	NewTokenHandle = HANDLE(INVALID_HANDLE_VALUE)
	_OpenProcessToken(ProcessHandle, DesiredAccess, byref(NewTokenHandle))
	return NewTokenHandle.value

def _internal_GetTokenInformation(hTokenHandle, TokenInformationClass, TokenInformation):
	_GetTokenInformation = windll.advapi32.GetTokenInformation
	_GetTokenInformation.argtypes = [HANDLE, TOKEN_INFORMATION_CLASS, LPVOID, DWORD, PDWORD]
	_GetTokenInformation.restype  = bool
	_GetTokenInformation.errcheck = RaiseIfZero

	ReturnLength = DWORD(0)
	TokenInformationLength = SIZEOF(TokenInformation)
	_GetTokenInformation(hTokenHandle, TokenInformationClass, byref(TokenInformation), TokenInformationLength, byref(ReturnLength))
	if ReturnLength.value != TokenInformationLength:
		raise ctypes.WinError(ERROR_INSUFFICIENT_BUFFER)
	return TokenInformation

# typedef enum tagTOKEN_TYPE {
#   TokenPrimary         = 1,
#   TokenImpersonation
# } TOKEN_TYPE, *PTOKEN_TYPE;

TOKEN_TYPE = ctypes.c_int
PTOKEN_TYPE = POINTER(TOKEN_TYPE)


# typedef struct _LUID {
#   DWORD LowPart;
#   LONG HighPart;
# } LUID,
#  *PLUID;
class LUID(Structure):
	_fields_ = [
		("LowPart",     DWORD),
		("HighPart",    LONG),
	]

PLUID = POINTER(LUID)

# typedef enum _SECURITY_IMPERSONATION_LEVEL {
#   SecurityAnonymous,
#   SecurityIdentification,
#   SecurityImpersonation,
#   SecurityDelegation
# } SECURITY_IMPERSONATION_LEVEL, *PSECURITY_IMPERSONATION_LEVEL;

SecurityAnonymous       = 0
SecurityIdentification  = 1
SecurityImpersonation   = 2
SecurityDelegation      = 3

SECURITY_IMPERSONATION_LEVEL = ctypes.c_int
PSECURITY_IMPERSONATION_LEVEL = POINTER(SECURITY_IMPERSONATION_LEVEL)


# typedef struct _TOKEN_STATISTICS {
#   LUID                         TokenId;
#   LUID                         AuthenticationId;
#   LARGE_INTEGER                ExpirationTime;
#   TOKEN_TYPE                   TokenType;
#   SECURITY_IMPERSONATION_LEVEL ImpersonationLevel;
#   DWORD                        DynamicCharged;
#   DWORD                        DynamicAvailable;
#   DWORD                        GroupCount;
#   DWORD                        PrivilegeCount;
#   LUID                         ModifiedId;
# } TOKEN_STATISTICS, *PTOKEN_STATISTICS;
class TOKEN_STATISTICS(Structure):
	_fields_ = [
		("TokenId",             LUID),
		("AuthenticationId",    LUID),
		("ExpirationTime",      LONGLONG),  # LARGE_INTEGER
		("TokenType",           TOKEN_TYPE),
		("ImpersonationLevel",  SECURITY_IMPERSONATION_LEVEL),
		("DynamicCharged",      DWORD),
		("DynamicAvailable",    DWORD),
		("GroupCount",          DWORD),
		("PrivilegeCount",      DWORD),
		("ModifiedId",          LUID),
	]
PTOKEN_STATISTICS = POINTER(TOKEN_STATISTICS)

# BOOL WINAPI GetTokenInformation(
#   __in       HANDLE TokenHandle,
#   __in       TOKEN_INFORMATION_CLASS TokenInformationClass,
#   __out_opt  LPVOID TokenInformation,
#   __in       DWORD TokenInformationLength,
#   __out      PDWORD ReturnLength
# );
def GetTokenInformation(hTokenHandle, TokenInformationClass):
	if TokenInformationClass <= 0 or TokenInformationClass > MaxTokenInfoClass:
		raise ValueError("Invalid value for TokenInformationClass (%i)" % TokenInformationClass)

	# Token statistics.
	if TokenInformationClass == TokenStatistics:
		TokenInformation = TOKEN_STATISTICS()
		_internal_GetTokenInformation(hTokenHandle, TokenInformationClass, TokenInformation)
		return TokenInformation # TODO add a class wrapper?

class LSA_LAST_INTER_LOGON_INFO(Structure):
	_fields_ = [
		("LastSuccessfulLogon", LONGLONG),
		("LastFailedLogon", LONGLONG),
		("FailedAttemptCountSinceLastSuccessfulLogon", ULONG),
	]
PLSA_LAST_INTER_LOGON_INFO = POINTER(LSA_LAST_INTER_LOGON_INFO)

class SECURITY_LOGON_SESSION_DATA(Structure):
	_fields_ = [
		("Size", ULONG),
		("LogonId", LUID),
		("UserName", UNICODE_STRING),
		("LogonDomain", UNICODE_STRING),
		("AuthenticationPackage", UNICODE_STRING),
		("Session", ULONG),
		("Sid", PSID),
		("LogonTime", LONGLONG),
		("LogonServer", UNICODE_STRING),
		("DnsDomainName", UNICODE_STRING),
		("Upn", UNICODE_STRING),
		("UserFlags", ULONG),
		("LastLogonInfo", LSA_LAST_INTER_LOGON_INFO),
		("LogonScript", UNICODE_STRING),
		("ProfilePath", UNICODE_STRING),
		("HomeDirectory", UNICODE_STRING),
		("HomeDirectoryDrive", UNICODE_STRING),
		("LogoffTime", LONGLONG),
		("KickOffTime", LONGLONG),
		("PasswordLastSet", LONGLONG),
		("PasswordCanChange", LONGLONG),
		("PasswordMustChange", LONGLONG),
	]
PSECURITY_LOGON_SESSION_DATA = POINTER(SECURITY_LOGON_SESSION_DATA)

def LsaGetLogonSessionData(luid):
	_LsaGetLogonSessionData = windll.Secur32.LsaGetLogonSessionData
	_LsaGetLogonSessionData.argtypes = [PLUID, POINTER(PSECURITY_LOGON_SESSION_DATA)]
	_LsaGetLogonSessionData.restype  = DWORD
	#_LsaGetLogonSessionData.errcheck = RaiseIfNotErrorSuccess

	secdata = PSECURITY_LOGON_SESSION_DATA()

	x = _LsaGetLogonSessionData(luid, byref(secdata))
	res = {
		'username' : secdata.contents.UserName.get_string(),
		'usersid' : ConvertSidToStringSidW(secdata.contents.Sid, True),
		'domain' : secdata.contents.LogonDomain.get_string(),
		'logonserver' : secdata.contents.LogonServer.get_string(),
		'dnsdomainname' : secdata.contents.DnsDomainName.get_string(),
	}
	return res

# DWORD WINAPI GetCurrentProcessId(void);
def GetCurrentProcessId():
	_GetCurrentProcessId = windll.kernel32.GetCurrentProcessId
	_GetCurrentProcessId.argtypes = []
	_GetCurrentProcessId.restype  = DWORD
	return _GetCurrentProcessId()

# HANDLE WINAPI OpenProcess(
#   __in  DWORD dwDesiredAccess,
#   __in  BOOL bInheritHandle,
#   __in  DWORD dwProcessId
# );
def OpenProcess(dwDesiredAccess, bInheritHandle, dwProcessId):
	_OpenProcess = windll.kernel32.OpenProcess
	_OpenProcess.argtypes = [DWORD, BOOL, DWORD]
	_OpenProcess.restype  = HANDLE

	if dwDesiredAccess is None:
		dwDesiredAccess = 0x10000000
	hProcess = _OpenProcess(dwDesiredAccess, bool(bInheritHandle), dwProcessId)
	if hProcess == NULL:
		raise ctypes.WinError()
	return hProcess



import io
from winacl.dtyp.wcee import keyblobselector

class CryptoAPIKeyProperty:
	def __init__(self):
		self.length = None
		self.type = None
		self.unknown1 = None
		self.name_len = None
		self.name = None
		self.value_len = None
		self.value = None
	
	@staticmethod
	def from_bytes(data:bytes):
		return CryptoAPIKeyProperty.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		prop = CryptoAPIKeyProperty()
		prop.length = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		prop.type = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		prop.unknown1 = buff.read(4)
		prop.name_len = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		prop.value_len = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		prop.name = buff.read(prop.name_len).decode('utf-16-le')
		prop.value = buff.read(prop.value_len)
		if prop.name == 'NgcSoftwareKeyPbkdf2Round':
			prop.value = int.from_bytes(prop.value, byteorder='little', signed=False)
		return prop
	
	def __str__(self):
		t = '= CryptoAPIKeyProperty =\n'
		t += 'Length: %s\n' % self.length
		t += 'Type: %s\n' % self.type
		t += 'Unknown1: %s\n' % self.unknown1
		t += 'Name Length: %s\n' % self.name_len
		t += 'Name: %s\n' % self.name
		t += 'Value Length: %s\n' % self.value_len
		t += 'Value: %s\n' % self.value
		return t

class CryptoAPIKeyProperties:
	# after DPAPI decryption, the properties are stored in a list of CryptoAPIKeyProperty objects
	def __init__(self):
		self.properties = {}
	
	@staticmethod
	def from_bytes(data:bytes):
		return CryptoAPIKeyProperties.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		prop = CryptoAPIKeyProperties()
		while True:
			p = CryptoAPIKeyProperty.from_buffer(buff)
			prop.properties[p.name] = p
			if buff.tell() == len(buff.getbuffer()):
				return prop.properties


class CryptoAPIPublicKey:
	def __init__(self):
		self.header_size = None
		self.unknown1 = None
		self.unknown2 = None
		self.header1_size = None
		self.header2_size = None
		self.header1 = None
		self.header2 = None
		self.keyblob = None
	
	@staticmethod
	def from_bytes(data:bytes):
		return CryptoAPIPublicKey.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff:io.BytesIO):
		start = buff.tell()
		key = CryptoAPIPublicKey()
		key.header_size = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		key.unknown1 = buff.read(4)
		key.unknown2 = buff.read(4)
		key.header1_size = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		key.header2_size = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		key.header1 = buff.read(key.header1_size).decode('utf-16-le')
		key.header2 = buff.read(key.header2_size)
		buff.read(20)

		# this point it's a public key.
		keystruct = keyblobselector(buff)
		key.keyblob = keystruct.from_buffer(buff)
		return key

	def __str__(self):
		t = '== CryptoAPIPublicKey ==\n'
		t += 'Header Size: %s\n' % self.header_size
		t += 'Unknown1: %s\n' % self.unknown1
		t += 'Unknown2: %s\n' % self.unknown2
		t += 'Header1 Size: %s\n' % self.header1_size
		t += 'Header2 Size: %s\n' % self.header2_size
		t += 'Header1: %s\n' % self.header1
		t += 'Header2: %s\n' % self.header2

		t += 'Key: %s\n' % self.keyblob
		return t


class CryptoAPIKeyFile:
	def __init__(self):
		self.type = None
		self.unknown1 = None
		self.description_len = None
		self.unknown2 = None
		self.field_cnt = None
		self.filed_lengths = []
		self.description = None
		self.fields = []
	
	def from_bytes(data:bytes):
		return CryptoAPIKeyFile.from_buffer(io.BytesIO(data))

	def from_buffer(buff:io.BytesIO):
		key = CryptoAPIKeyFile()
		key.type = buff.read(4)
		key.unknown1 = buff.read(4)
		key.description_len = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		key.unknown2 = buff.read(2)
		key.field_cnt = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		for _ in range(0,key.field_cnt):
			key.filed_lengths.append(int.from_bytes(buff.read(4), byteorder='little', signed=False))
		buff.seek(44,0)
		key.description = buff.read(key.description_len).decode('utf-16-le')
		for i, size in enumerate(key.filed_lengths):
			field = buff.read(size)
			if i == 0:
				field = CryptoAPIPublicKey.from_bytes(field) # other fields are -hopefully- DPAPI blobs
			key.fields.append(field)
		return key
	
	def __str__(self):
		t = '== CryptoAPIKeyDescriptor ==\n'
		t += 'Type: %s\n' % self.type
		t += 'Unknown1: %s\n' % self.unknown1
		t += 'Description Length: %s\n' % self.description_len
		t += 'Unknown2: %s\n' % self.unknown2
		t += 'Field Count: %s\n' % self.field_cnt
		t += 'Description: %s\n' % self.description
		for i in range(self.field_cnt):
			t += 'Field %d: %s\n' % (i, self.fields[i])
		return t


def main():
	with open('/home/webdev/Desktop/winhellotest/winhello_test/cryptokeys/5b64a9701545a917d7778d1d19fd2d78_50962b6d-f3c6-40f5-b6ab-81e1beb9be3a', 'rb') as f:
		data = f.read()
	
	key = CryptoAPIKeyFile.from_bytes(data)
	print(str(key))

if __name__ == '__main__':
	main()



import io
import math

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-wcce/5cf2e6b9-3195-4f85-bc18-05b50e6d4e11
class RSAPrivateKeyBlob:
	def __init__(self):
		self.type = None
		self.version = None
		self.reserved = None
		self.keyalg = None
		self.magic = None
		self.bitlen = None
		self.pubexp = None
		self.modulus = None
		self.p = None
		self.q = None
		self.dp = None
		self.dq = None
		self.iq = None
		self.d = None

	@staticmethod
	def from_bytes(data):
		return RSAPrivateKeyBlob.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		pk = RSAPrivateKeyBlob()
		pk.type = buff.read(1)[0]
		pk.version = buff.read(1)[0]
		pk.reserved = int.from_bytes(buff.read(2), byteorder='little', signed=False)
		pk.keyalg = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pk.magic = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pk.bitlen = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		c8 = int(math.ceil(pk.bitlen/8))
		c16 = int(math.ceil(pk.bitlen/16))
		pk.pubexp = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pk.modulus = int.from_bytes(buff.read(c8), byteorder='little', signed=False)
		pk.p = int.from_bytes(buff.read(c16), byteorder='little', signed=False)
		pk.q = int.from_bytes(buff.read(c16), byteorder='little', signed=False)
		pk.dp = int.from_bytes(buff.read(c16), byteorder='little', signed=False)
		pk.dq = int.from_bytes(buff.read(c16), byteorder='little', signed=False)
		pk.iq = int.from_bytes(buff.read(c16), byteorder='little', signed=False)
		pk.d = int.from_bytes(buff.read(c8), byteorder='little', signed=False)
		return pk

	def get_key(self):
		public_numbers = rsa.RSAPublicNumbers(self.pubexp, self.modulus)
		numbers = rsa.RSAPrivateNumbers(self.p, self.q, self.d, self.dp, self.dq, self.iq, public_numbers)
		return default_backend().load_rsa_private_numbers(numbers, unsafe_skip_rsa_key_validation=False)

	def get_pubkey(self):
		public_numbers = rsa.RSAPublicNumbers(self.pubexp, self.modulus)
		return default_backend().load_rsa_public_numbers(public_numbers)

	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
		return t

# https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-wcce/540b7b8b-2232-45c8-9d7c-af7a5d5218ed
class BCRYPTRSAKeyBlob:
	def __init__(self):
		self.magic = None
		self.bitlen = None
		self.pubexplen = None
		self.moduluslen = None
		self.plen = None
		self.qlen = None
		self.pubexp = None
		self.modulus = None
		self.p = None
		self.q = None

	@staticmethod
	def from_bytes(data):
		return BCRYPTRSAKeyBlob.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		pk = BCRYPTRSAKeyBlob()
		pk.magic = buff.read(4)
		pk.bitlen = int.from_bytes(buff.read(4), 'little', signed=False)
		pk.pubexplen = int.from_bytes(buff.read(4), 'little', signed=False)
		pk.moduluslen = int.from_bytes(buff.read(4), 'little', signed=False)
		pk.plen = int.from_bytes(buff.read(4), 'little', signed=False)
		pk.qlen = int.from_bytes(buff.read(4), 'little', signed=False)
		pk.pubexp = int.from_bytes(buff.read(pk.pubexplen), 'big', signed=False)
		pk.modulus = int.from_bytes(buff.read(pk.moduluslen), 'big', signed=False)
		pk.p = int.from_bytes(buff.read(pk.plen), 'big', signed=False)
		pk.q = int.from_bytes(buff.read(pk.qlen), 'big', signed=False)
		return pk

	def get_key(self):
		phi = (self.p - 1) * (self.q - 1)
		d = rsa._modinv(self.pubexp, phi)
		dmp1 = rsa.rsa_crt_dmp1(d, self.p)
		dmq1 = rsa.rsa_crt_dmq1(d, self.q)
		iqmp = rsa.rsa_crt_iqmp(self.p, self.q)
		numbers = rsa.RSAPrivateNumbers(
			p=self.p,
			q=self.q,
			d=d,
			dmp1=dmp1,
			dmq1=dmq1,
			iqmp=iqmp,
			public_numbers=rsa.RSAPublicNumbers(
				e=self.pubexp,
				n=self.modulus
			)
		)
		return default_backend().load_rsa_private_numbers(numbers, unsafe_skip_rsa_key_validation=False)

	def get_pubkey(self):
		public_numbers = rsa.RSAPublicNumbers(self.pubexp, self.modulus)
		return default_backend().load_rsa_public_numbers(public_numbers)

	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
		return t


from winacl.dtyp.acl import ACL
from winacl.dtyp.sid import SID
import enum
import io

class SE_SACL(enum.IntFlag):
	SE_DACL_AUTO_INHERIT_REQ = 0x0100 	#Indicates a required security descriptor in which the discretionary access control list (DACL) is set up to support automatic propagation of inheritable access control entries (ACEs) to existing child objects.
										#For access control lists (ACLs) that support auto inheritance, this bit is always set. Protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function to convert a security descriptor and set this flag.
	SE_DACL_AUTO_INHERITED = 0x0400     #Indicates a security descriptor in which the discretionary access control list (DACL) is set up to support automatic propagation of inheritable access control entries (ACEs) to existing child objects.
										#For access control lists (ACLs) that support auto inheritance, this bit is always set. Protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function to convert a security descriptor and set this flag.
	SE_DACL_DEFAULTED = 0x0008			#Indicates a security descriptor with a default DACL. For example, if the creator an object does not specify a DACL, the object receives the default DACL from the access token of the creator. This flag can affect how the system treats the DACL with respect to ACE inheritance. The system ignores this flag if the SE_DACL_PRESENT flag is not set.
										#This flag is used to determine how the final DACL on the object is to be computed and is not stored physically in the security descriptor control of the securable object.
										#To set this flag, use the SetSecurityDescriptorDacl function.
	SE_DACL_PRESENT = 0x0004			#Indicates a security descriptor that has a DACL. If this flag is not set, or if this flag is set and the DACL is NULL, the security descriptor allows full access to everyone.
										#This flag is used to hold the security information specified by a caller until the security descriptor is associated with a securable object. After the security descriptor is associated with a securable object, the SE_DACL_PRESENT flag is always set in the security descriptor control.
										#To set this flag, use the SetSecurityDescriptorDacl function.
	SE_DACL_PROTECTED = 0x1000			#Prevents the DACL of the security descriptor from being modified by inheritable ACEs. To set this flag, use the SetSecurityDescriptorControl function.
	SE_GROUP_DEFAULTED = 0x0002			#Indicates that the security identifier (SID) of the security descriptor group was provided by a default mechanism. This flag can be used by a resource manager to identify objects whose security descriptor group was set by a default mechanism. To set this flag, use the SetSecurityDescriptorGroup function.
	SE_OWNER_DEFAULTED = 0x0001			#Indicates that the SID of the owner of the security descriptor was provided by a default mechanism. This flag can be used by a resource manager to identify objects whose owner was set by a default mechanism. To set this flag, use the SetSecurityDescriptorOwner function.
	SE_RM_CONTROL_VALID = 0x4000		#Indicates that the resource manager control is valid.
	SE_SACL_AUTO_INHERIT_REQ = 0x0200	#Indicates a required security descriptor in which the system access control list (SACL) is set up to support automatic propagation of inheritable ACEs to existing child objects.
										#The system sets this bit when it performs the automatic inheritance algorithm for the object and its existing child objects. To convert a security descriptor and set this flag, protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function.
	SE_SACL_AUTO_INHERITED = 0x0800		#Indicates a security descriptor in which the system access control list (SACL) is set up to support automatic propagation of inheritable ACEs to existing child objects.
										#The system sets this bit when it performs the automatic inheritance algorithm for the object and its existing child objects. To convert a security descriptor and set this flag, protected servers can call the ConvertToAutoInheritPrivateObjectSecurity function.
	SE_SACL_DEFAULTED = 0x0008			#A default mechanism, rather than the original provider of the security descriptor, provided the SACL. This flag can affect how the system treats the SACL, with respect to ACE inheritance. The system ignores this flag if the SE_SACL_PRESENT flag is not set. To set this flag, use the SetSecurityDescriptorSacl function.
	SE_SACL_PRESENT   = 0x0010			#Indicates a security descriptor that has a SACL. To set this flag, use the SetSecurityDescriptorSacl function.
	SE_SACL_PROTECTED = 0x2000			#Prevents the SACL of the security descriptor from being modified by inheritable ACEs. To set this flag, use the SetSecurityDescriptorControl function.
	SE_SELF_RELATIVE  = 0x8000			#Indicates a self-relative security descriptor. If this flag is not set, the security descriptor is in absolute format. For more information, see Absolute and Self-Relative Security Descriptors.

sddl_acl_control_flags = {
	"P"  : SE_SACL.SE_DACL_PROTECTED,
	"AR" : SE_SACL.SE_DACL_AUTO_INHERIT_REQ,
	"AI" : SE_SACL.SE_DACL_AUTO_INHERITED,
	"SR" : SE_SACL.SE_SELF_RELATIVE,
	#"NO_ACCESS_CONTROL" : 0
}
sddl_acl_control_flags_inv = {v: k for k, v in sddl_acl_control_flags.items()}

def sddl_acl_control(flags):
	t = ''
	for x in sddl_acl_control_flags_inv:
		if x == SE_SACL.SE_SELF_RELATIVE:
			continue #this flag is always set implicitly
		if x in flags:
			t += sddl_acl_control_flags_inv[x]
	return t

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/7d4dac05-9cef-4563-a058-f108abecce1d
class SECURITY_DESCRIPTOR:
	def __init__(self, object_type = None):
		self.Revision:int = 1
		self.Sbz1:int = 0 #default value but SDDL doesnt store this info and in some cases this field is nonzero
		self.Control:SE_SACL = None
		self.Owner:SID = None
		self.Group:SID = None
		self.Sacl:ACL = None
		self.Dacl:ACL = None

		self.object_type = object_type #high level info, not part of the struct
	
	@staticmethod
	def from_bytes(data, object_type = None):
		return SECURITY_DESCRIPTOR.from_buffer(io.BytesIO(data), object_type)
	
	def to_bytes(self):
		buff = io.BytesIO()
		self.to_buffer(buff)
		buff.seek(0)
		return buff.read()

	def to_buffer(self, buff):
		start = buff.tell()
		buff_data = io.BytesIO()
		OffsetOwner = 0
		OffsetGroup = 0
		OffsetSacl = 0
		OffsetDacl = 0

		if self.Owner is not None:
			buff_data.write(self.Owner.to_bytes())
			OffsetOwner = start + 20
		
		if self.Group is not None:
			OffsetGroup = start + 20 + buff_data.tell()
			buff_data.write(self.Group.to_bytes())
			
		
		if self.Sacl is not None:
			OffsetSacl = start + 20 + buff_data.tell()
			buff_data.write(self.Sacl.to_bytes())
			
		
		if self.Dacl is not None:
			OffsetDacl = start + 20 + buff_data.tell()
			buff_data.write(self.Dacl.to_bytes())
			

		
		buff.write(self.Revision.to_bytes(1, 'little', signed = False))
		buff.write(self.Sbz1.to_bytes(1, 'little', signed = False))
		buff.write(self.Control.to_bytes(2, 'little', signed = False))
		buff.write(OffsetOwner.to_bytes(4, 'little', signed = False))
		buff.write(OffsetGroup.to_bytes(4, 'little', signed = False))
		buff.write(OffsetSacl.to_bytes(4, 'little', signed = False))
		buff.write(OffsetDacl.to_bytes(4, 'little', signed = False))
		buff_data.seek(0)
		buff.write(buff_data.read())


	@staticmethod
	def from_buffer(buff, object_type = None):
		sd = SECURITY_DESCRIPTOR(object_type)
		sd.Revision = int.from_bytes(buff.read(1), 'little', signed = False)
		sd.Sbz1 =  int.from_bytes(buff.read(1), 'little', signed = False)
		sd.Control = SE_SACL(int.from_bytes(buff.read(2), 'little', signed = False))
		OffsetOwner  = int.from_bytes(buff.read(4), 'little', signed = False)
		OffsetGroup  = int.from_bytes(buff.read(4), 'little', signed = False)
		OffsetSacl  = int.from_bytes(buff.read(4), 'little', signed = False)
		OffsetDacl  = int.from_bytes(buff.read(4), 'little', signed = False)

		if OffsetOwner > 0:
			buff.seek(OffsetOwner)
			sd.Owner = SID.from_buffer(buff)
		
		if OffsetGroup > 0:
			buff.seek(OffsetGroup)
			sd.Group = SID.from_buffer(buff)
			
		if OffsetSacl > 0:
			buff.seek(OffsetSacl)
			sd.Sacl = ACL.from_buffer(buff, object_type)
		
		if OffsetDacl > 0:
			buff.seek(OffsetDacl)
			sd.Dacl = ACL.from_buffer(buff, object_type)
			
		return sd
	
	def to_sddl(self, object_type = None):
		t=''
		if self.Owner is not None:
			t +=  'O:' + self.Owner.to_sddl()
		if self.Group is not None:
			t += 'G:' + self.Group.to_sddl()
		if self.Sacl is not None:
			t+= 'S:' + sddl_acl_control(self.Control) + self.Sacl.to_sddl(object_type)
		if self.Dacl is not None:
			t+= 'D:' + sddl_acl_control(self.Control) + self.Dacl.to_sddl(object_type)
		return t

	@staticmethod
	def from_sddl(sddl:str, object_type = None, domain_sid = None):
		sd = SECURITY_DESCRIPTOR(object_type = object_type)
		params = sddl.split(':')
		np = [params[0]]
		i = 1
		while i < len(params):
			np.append(params[i][:-1])
			np.append(params[i][-1])
			i += 1
		params = {}
		i = 0 
		while i < len(np):
			if np[i] == ')':
				break
			params[np[i]] = np[i+1]
			i += 2
		
		
		sd.Control = SE_SACL.SE_SELF_RELATIVE
		fk = None
		if 'D' in params:
			fk = 'D'
		elif 'S' in params:
			fk = 'S'
		
		if fk is not None:
			if '(' in params[fk]:
				flags, acl = params[fk].split('(', 1)
			else:
				flags = params[fk]
			if flags.upper().find('P') != -1:
				sd.Control |= SE_SACL.SE_DACL_PROTECTED
				sd.Control |= SE_SACL.SE_SACL_PROTECTED
				flags = flags.replace('P', '')
			for _ in range(len(flags)):
				x = flags[:2]
				cf = sddl_acl_control_flags[x]
				if cf == SE_SACL.SE_DACL_AUTO_INHERIT_REQ:
					sd.Control |= SE_SACL.SE_DACL_AUTO_INHERIT_REQ
					sd.Control |= SE_SACL.SE_SACL_AUTO_INHERIT_REQ
				elif cf == SE_SACL.SE_DACL_AUTO_INHERITED:
					sd.Control |= SE_SACL.SE_DACL_AUTO_INHERITED
					sd.Control |= SE_SACL.SE_SACL_AUTO_INHERITED
				else:
					sd.Control |= cf
				
				flags = flags[2:]
				if flags == '':
					break

		
		if 'O' in params:
			sd.Owner = SID.from_sddl(params['O'], domain_sid = domain_sid)
		if 'G' in params:
			sd.Group = SID.from_sddl(params['G'], domain_sid = domain_sid)
		if 'D' in params:
			sd.Control |= SE_SACL.SE_DACL_PRESENT
			acl = params['D']
			m = acl.find('(')
			if m != -1:
				sd.Dacl = ACL.from_sddl(acl[m:], object_type=object_type, domain_sid = domain_sid)
		if 'S' in params:
			sd.Control |= SE_SACL.SE_SACL_PRESENT
			acl = params['S']
			m = acl.find('(')
			if m != -1:
				sd.Sacl = ACL.from_sddl(acl[m:], object_type=object_type, domain_sid = domain_sid)
			
		return sd
			
	def __str__(self):
		t = '=== SECURITY_DESCRIPTOR ==\r\n'
		t+= 'Revision : %s\r\n' % self.Revision
		t+= 'Control : %s\r\n' % self.Control
		t+= 'Owner : %s\r\n' % self.Owner
		t+= 'Group : %s\r\n' % self.Group
		t+= 'Sacl : %s\r\n' % self.Sacl
		t+= 'Dacl : %s\r\n' % self.Dacl
		return t

	def __eq__(self, sd):
		if not isinstance(sd, SECURITY_DESCRIPTOR):
			return False
		if sd.Revision != self.Revision:
			return False
		if sd.Sbz1 != self.Sbz1:
			return False
		if self.Control != self.Control:
			return False
		if self.Owner != self.Owner:
			return False
		if self.Group != self.Group:
			return False
		if sd.Sacl is not None and self.Sacl is None:
			return False
		if sd.Dacl is not None and self.Dacl is None:
			return False
		if sd.Sacl is None and self.Sacl is not None:
			return False
		if sd.Dacl is None and self.Dacl is not None:
			return False
		
		if self.Sacl is not None:
			if self.Sacl != sd.Sacl:
				return False
		if self.Dacl is not None:
			if self.Dacl != sd.Dacl:
				return False

		return True
	
	def diff(self, sd):
		diff_res = {}
		if sd.Revision != self.Revision:
			diff_res['revision'] = [self.Revision, sd.Revision]
		if sd.Sbz1 != self.Sbz1:
			diff_res['Sbz1'] = [self.Sbz1, sd.Sbz1]
		if self.Control != self.Control:
			diff_res['Control'] = [self.Control, sd.Control]
		if self.Owner != self.Owner:
			diff_res['Owner'] = [self.Owner, sd.Owner]
		if self.Group != self.Group:
			diff_res['Group'] = [self.Group, sd.Group]
		if sd.Sacl is not None and self.Sacl is None:
			diff_res['Sacl_added'] = None
		if sd.Dacl is not None and self.Dacl is None:
			diff_res['Dacl_added'] = None
		if sd.Sacl is None and self.Sacl is not None:
			diff_res['Sacl_deleted'] = None
		if sd.Dacl is None and self.Dacl is not None:
			diff_res['Dacl_deleted'] = None
		
		if self.Sacl is not None and sd.Sacl is not None:
			sacl_diff = self.Sacl.diff(sd.Sacl)
			if len(sacl_diff) > 0:
				diff_res['Sacl'] = sacl_diff
		if self.Dacl is not None:
			dacl_diff = self.Dacl.diff(sd.Dacl)
			if len(dacl_diff) > 0:
				diff_res['Dacl'] = dacl_diff

		return diff_res
	
if __name__ == '__main__':
	ds = 'S-1-5-21-3448413973-1765323015-1500960949'
	x = 'O:BAG:DUD:AI(A;OICI;FA;;;DA)(A;OICIID;FA;;;SY)(A;OICIID;FA;;;BA)(A;OICIID;0x1200a9;;;BU)(A;CIID;0x100004;;;BU)(A;CIID;0x100002;;;BU)(A;OICIIOID;FA;;;CO)(A;OICIID;FA;;;DA)'
	a = SECURITY_DESCRIPTOR.from_sddl(x, domain_sid = ds)
	print(x)
	print(a.to_sddl())
	print(x == a.to_sddl())

	print(a)




from winacl.functions.defines import *
import enum
import ctypes
from winacl.functions.kernel32 import GetLastError, LocalFree, READ_CONTROL

class DnsConfig(enum.Enum):
	PrimaryDomainName_W = 0
	PrimaryDomainName_A = 1
	PrimaryDomainName_UTF8 = 2
	AdapterDomainName_W = 3
	AdapterDomainName_A = 4
	AdapterDomainName_UTF8 = 5
	DnsServerList = 6
	SearchList = 7
	AdapterInfo = 8
	PrimaryHostNameRegistrationEnabled = 9
	AdapterHostNameRegistrationEnabled = 10
	AddressRegistrationMaxCount = 11
	HostName_W = 12
	HostName_A = 13
	HostName_UTF8 = 14
	FullHostName_W = 15
	FullHostName_A = 16
	FullHostName_UTF8 = 17
	NameServer = 18

def DnsQueryConfig(dnsconfig, adaptername):
	_DnsQueryConfig = windll.Dnsapi.DnsQueryConfig
	_DnsQueryConfig.argtypes = [DWORD, DWORD, PVOID, PVOID, PVOID, PDWORD]
	_DnsQueryConfig.restype  = DWORD
	_DnsQueryConfig.errcheck = RaiseIfNotErrorSuccess

	if isinstance(adaptername, str):
		padaptername = ctypes.create_unicode_buffer(adaptername)

	bufptr = ctypes.create_string_buffer(1024)
	buflenptr = ctypes.pointer(DWORD(1024))

	_DnsQueryConfig(dnsconfig.value, 0, padaptername, None, byref(bufptr), buflenptr)

	data = ctypes.string_at(bufptr, buflenptr.contents.value)

	return data

def DnsQueryConfig_ServerList(adaptername):
	data = DnsQueryConfig(DnsConfig.DnsServerList, adaptername)





# list services current user has write access to
# list services given user has write access to
# list sd of services
# list services
# get sd of given service

from winacl.functions.highlevel import *
from winacl.dtyp.ace import SERVICE_ACCESS_MASK

def main():
	import argparse
	parser = argparse.ArgumentParser(description='Pure Python implementation of Mimikatz --and more--')
	parser.add_argument('-v', '--verbose', action='count', default=0)

	subparsers = parser.add_subparsers(help = 'commands')
	subparsers.required = True
	subparsers.dest = 'cmdtype'

	services_group = subparsers.add_parser('srv', help='Service related commands')
	services_group_sub = services_group.add_subparsers()
	services_group_sub.required = True
	services_group_sub.dest = 'command'

	srv_ls = services_group_sub.add_parser('ls', help='listing services')
	srv_ls.add_argument('--user', help = 'user to use in eff mode. If omitted, the current user will be used. Format: "domain\\username" for domain users or "username" for local users ')
	srv_ls.add_argument('--sid', help = 'SID of the user to use in eff mode.')
	srv_ls.add_argument('--sddl', action = 'store_true', help = 'Security Descriptors will be listed in SDDL format')
	srv_ls.add_argument('--eff', action = 'store_true', help = 'Effective permissions will be printed insted of Security Descriptors')

	srv_name = services_group_sub.add_parser('name', help='operate on specific service')
	srv_name.add_argument('service_name', help = 'Name of the service to operate on (short)')
	srv_name.add_argument('--user', help = 'user to use in eff mode. If omitted, the current user will be used. Format: "domain\\username" for domain users or "username" for local users ')
	srv_name.add_argument('--sid', help = 'SID of the user to use in eff mode.')
	srv_name.add_argument('--sddl', action = 'store_true', help = 'Security Descriptors will be listed in SDDL format')
	srv_name.add_argument('--eff', action = 'store_true', help = 'Effective permissions will be printed insted of Security Descriptors')


	args = parser.parse_args()
	print(args)

	if args.cmdtype == 'srv':
		if args.command == 'ls':
			for srv_name, srv_sd in enumerate_all_service_sd():
				output = 'Name: %s ' % srv_name
				
				if isinstance(srv_sd, str):
					output += 'Error getting SD!'
					print(output)
					continue
			
				if args.sddl is True:
					output += 'SDDL: %s' % srv_sd.to_sddl()
				elif args.eff is True:
					if args.user is not None:
						max_perm = get_maximum_permissions_for_user(srv_sd, args.user)
					elif args.sid is not None:
						max_perm = GetEffectiveRightsFromAclW(srv_sd.Dacl, SID.from_string(args.sid))
					else:
						raise NotImplementedError()
					output += 'max_perm : %s value: %s' % (SERVICE_ACCESS_MASK(max_perm) , hex(max_perm))
				else:
					output += 'SD: %s' % str(srv_sd)
		
				print(output)
		elif args.command == 'name':
			srv_sd = get_service_sd(args.service_name)
			output = ''
			if args.sddl is True:
				print('SDDL: %s' % srv_sd.to_sddl())
			
			elif args.eff is True:
				if args.user is not None:
					max_perm = get_maximum_permissions_for_user(srv_sd, args.user)
				elif args.sid is not None:
					max_perm = GetEffectiveRightsFromAclW(srv_sd.Dacl, SID.from_string(args.sid))
				else:
					raise NotImplementedError()
				output += 'max_perm : %s value: %s' % (SERVICE_ACCESS_MASK(max_perm) , hex(max_perm))
			else:
				output += 'SD: %s' % str(srv_sd)
			
			print(output)
			

if __name__ == '__main__':
	main()


import io

# https://docs.microsoft.com/en-us/previous-versions/aa373931(v%3Dvs.80)
class GUID:
	def __init__(self):
		self.Data1:bytes = None
		self.Data2:bytes = None
		self.Data3:bytes = None
		self.Data4:bytes = None
	
	def to_bytes(self):
		return self.Data1[::-1] + self.Data2[::-1] + self.Data3[::-1] + self.Data4

	def to_buffer(self, buff):
		buff.write(self.to_bytes())

	@staticmethod
	def from_bytes(data):
		return GUID.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		guid = GUID()
		guid.Data1 = buff.read(4)[::-1]
		guid.Data2 = buff.read(2)[::-1]
		guid.Data3 = buff.read(2)[::-1]
		guid.Data4 = buff.read(8)
		return guid
		
	@staticmethod
	def from_string(str):
		guid = GUID()
		guid.Data1 = bytes.fromhex(str.split('-')[0])
		guid.Data2 = bytes.fromhex(str.split('-')[1])
		guid.Data3 = bytes.fromhex(str.split('-')[2])
		guid.Data4 = bytes.fromhex(str.split('-')[3])
		guid.Data4 += bytes.fromhex(str.split('-')[4])
		return guid			
		
	def __str__(self):
		return '-'.join([self.Data1.hex(), self.Data2.hex(),self.Data3.hex(),self.Data4[:2].hex(),self.Data4[2:].hex()])


import io
import enum
from typing import List

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/f992ad60-0fe4-4b87-9fed-beb478836861
class SID:
	def __init__(self):
		self.Revision:int = None
		self.SubAuthorityCount:int = None
		self.IdentifierAuthority:int = None
		self.SubAuthority:List[int] = []

		self.is_wellknown:bool = None
		self.wellknow_name:str = None
		
		self.wildcard:bool = None #this is for well-known-sid lookups
	
	@staticmethod
	def wellknown_sid_lookup(x):
		if x in well_known_sids_sid_name_map:
			return well_known_sids_sid_name_map[x]
		return False

	@staticmethod
	def wellknown_name_lookup(x):
		if x in well_known_sids_name_sid_map:
			return well_known_sids_name_sid_map[x]
		return False


	@staticmethod	
	def from_string(sid_str, wildcard = False):
		if sid_str[:4] != 'S-1-':
			raise Exception('This is not a SID')
		sid = SID()
		sid.wildcard = wildcard
		sid.Revision = 1
		sid_str = sid_str[4:]
		t = sid_str.split('-')[0]
		if t[:2] == '0x':
			sid.IdentifierAuthority = int(t[2:],16)
		else:
			sid.IdentifierAuthority = int(t)
			
		for p in sid_str.split('-')[1:]:
			try:
				p = int(p)
			except Exception as e:
				if wildcard != True:
					raise e
			sid.SubAuthority.append(p)
		return sid
		
	@staticmethod
	def from_bytes(data):
		return SID.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		sid = SID()
		sid.Revision = int.from_bytes(buff.read(1), 'little', signed = False)
		sid.SubAuthorityCount = int.from_bytes(buff.read(1), 'little', signed = False)
		sid.IdentifierAuthority = int.from_bytes(buff.read(6), 'big', signed = False)
		for _ in range(sid.SubAuthorityCount):
			sid.SubAuthority.append(int.from_bytes(buff.read(4), 'little', signed = False))
		return sid
		
	def to_bytes(self):
		t = self.Revision.to_bytes(1, 'little', signed = False)
		t += len(self.SubAuthority).to_bytes(1, 'little', signed = False)
		t += self.IdentifierAuthority.to_bytes(6, 'big', signed = False)
		for i in self.SubAuthority:
			t += i.to_bytes(4, 'little', signed = False)
		return t
		
	def __str__(self):
		t = 'S-1-'
		if self.IdentifierAuthority < 2**32:
			t += str(self.IdentifierAuthority)
		else:
			t += '0x' + self.IdentifierAuthority.to_bytes(6, 'big').hex().upper().rjust(12, '0')
		for i in self.SubAuthority:
			t += '-' + str(i)
		return t

	def __eq__(self, other):
		if isinstance(other, SID):
			return str(self) == str(other)
		elif isinstance(other, str):
			return str(self) == other
		return NotImplemented

	def to_sddl(self):
		x = str(self)
		return x
		# TODO: if we figure out how to properly convert the pretty names to the correct SIDs enable this part
		#       problem is that pretty names sometimes belong to full SIDs sometimes to RIDs only and there is no way to tell if it belonged to
		#       a domain-sid or a local sid
		
		#for val in sddl_val_name_map:
		#	if isinstance(val, str) is True and val == x:
		#		return sddl_val_name_map[val]
		#	elif isinstance(val, int) is True and self.SubAuthority[-1] == val:
		#		return sddl_val_name_map[val]
		#return x

	@staticmethod
	def from_sddl(sddl, domain_sid = None):
		if len(sddl) > 2:
			return SID.from_string(sddl)
		else:
			if sddl not in sddl_name_val_map:
				raise Exception('%s was not found in the well known sid definitions!' % sddl)
			account_sid_val = sddl_name_val_map[sddl]
			if isinstance(account_sid_val, str):
				return SID.from_string(account_sid_val)
			else:
				if domain_sid is None:
					raise Exception('Missing domain_sid! Cant convert "%s" to a SID' % sddl)
				return SID.from_string(domain_sid + '-' +str(account_sid_val))

# https://support.microsoft.com/en-us/help/243330/well-known-security-identifiers-in-windows-operating-systems
# https://docs.microsoft.com/en-us/windows/win32/secauthz/well-known-sids

well_known_sids_name_sid_map = {
	'NULL' : 'S-1-0-0',
	'EVERYONE' : 'S-1-1-0',
	'LOCAL' : 'S-1-2-0',
	'CONSOLE_LOGON' : 'S-1-2-1',
	'CREATOR_OWNER' : 'S-1-3-0',
	'CREATOR_GROUP' : 'S-1-3-1',
	'OWNER_SERVER' : 'S-1-3-2',
	'GROUP_SERVER' : 'S-1-3-3',
	'OWNER_RIGHTS' : 'S-1-3-4',
	'NT_AUTHORITY' : 'S-1-5',
	'DIALUP' : 'S-1-5-1',
	'NETWORK' : 'S-1-5-2',
	'BATCH' : 'S-1-5-3',
	'INTERACTIVE' : 'S-1-5-4',
	'SERVICE' : 'S-1-5-6',
	'ANONYMOUS' : 'S-1-5-7',
	'PROXY' : 'S-1-5-8',
	'ENTERPRISE_DOMAIN_CONTROLLERS' : 'S-1-5-9',
	'PRINCIPAL_SELF' : 'S-1-5-10',
	'AUTHENTICATED_USERS' : 'S-1-5-11',
	'RESTRICTED_CODE' : 'S-1-5-12',
	'TERMINAL_SERVER_USER' : 'S-1-5-13',
	'REMOTE_INTERACTIVE_LOGON' : 'S-1-5-14',
	'THIS_ORGANIZATION' : 'S-1-5-15',
	'IUSR' : 'S-1-5-17',
	'LOCAL_SYSTEM' : 'S-1-5-18',
	'LOCAL_SERVICE' : 'S-1-5-19',
	'NETWORK_SERVICE' : 'S-1-5-20',
	'COMPOUNDED_AUTHENTICATION' : 'S-1-5-21-0-0-0-496',
	'CLAIMS_VALID' : 'S-1-5-21-0-0-0-497',
	'BUILTIN_ADMINISTRATORS' : 'S-1-5-32-544',
	'BUILTIN_USERS' : 'S-1-5-32-545',
	'BUILTIN_GUESTS' : 'S-1-5-32-546',
	'POWER_USERS' : 'S-1-5-32-547',
	'ACCOUNT_OPERATORS' : 'S-1-5-32-548',
	'SERVER_OPERATORS' : 'S-1-5-32-549',
	'PRINTER_OPERATORS' : 'S-1-5-32-550',
	'BACKUP_OPERATORS' : 'S-1-5-32-551',
	'REPLICATOR' : 'S-1-5-32-552',
	'ALIAS_PREW2KCOMPACC' : 'S-1-5-32-554',
	'REMOTE_DESKTOP' : 'S-1-5-32-555',
	'NETWORK_CONFIGURATION_OPS' : 'S-1-5-32-556',
	'INCOMING_FOREST_TRUST_BUILDERS' : 'S-1-5-32-557',
	'PERFMON_USERS' : 'S-1-5-32-558',
	'PERFLOG_USERS' : 'S-1-5-32-559',
	'WINDOWS_AUTHORIZATION_ACCESS_GROUP' : 'S-1-5-32-560',
	'TERMINAL_SERVER_LICENSE_SERVERS' : 'S-1-5-32-561',
	'DISTRIBUTED_COM_USERS' : 'S-1-5-32-562',
	'IIS_IUSRS' : 'S-1-5-32-568',
	'CRYPTOGRAPHIC_OPERATORS' : 'S-1-5-32-569',
	'EVENT_LOG_READERS' : 'S-1-5-32-573',
	'CERTIFICATE_SERVICE_DCOM_ACCESS' : 'S-1-5-32-574',
	'RDS_REMOTE_ACCESS_SERVERS' : 'S-1-5-32-575',
	'RDS_ENDPOINT_SERVERS' : 'S-1-5-32-576',
	'RDS_MANAGEMENT_SERVERS' : 'S-1-5-32-577',
	'HYPER_V_ADMINS' : 'S-1-5-32-578',
	'ACCESS_CONTROL_ASSISTANCE_OPS' : 'S-1-5-32-579',
	'REMOTE_MANAGEMENT_USERS' : 'S-1-5-32-580',
	'WRITE_RESTRICTED_CODE' : 'S-1-5-33',
	'NTLM_AUTHENTICATION' : 'S-1-5-64-10',
	'SCHANNEL_AUTHENTICATION' : 'S-1-5-64-14',
	'DIGEST_AUTHENTICATION' : 'S-1-5-64-21',
	'THIS_ORGANIZATION_CERTIFICATE' : 'S-1-5-65-1',
	'NT_SERVICE' : 'S-1-5-80',
	'USER_MODE_DRIVERS' : 'S-1-5-84-0-0-0-0-0',
	'LOCAL_ACCOUNT' : 'S-1-5-113',
	'LOCAL_ACCOUNT_AND_MEMBER_OF_ADMINISTRATORS_GROUP' : 'S-1-5-114',
	'OTHER_ORGANIZATION' : 'S-1-5-1000',
	'ALL_APP_PACKAGES' : 'S-1-15-2-1',
	'ML_UNTRUSTED' : 'S-1-16-0',
	'ML_LOW' : 'S-1-16-4096',
	'ML_MEDIUM' : 'S-1-16-8192',
	'ML_MEDIUM_PLUS' : 'S-1-16-8448',
	'ML_HIGH' : 'S-1-16-12288',
	'ML_SYSTEM' : 'S-1-16-16384',
	'ML_PROTECTED_PROCESS' : 'S-1-16-20480',
	'ML_SECURE_PROCESS' : 'S-1-16-28672',
	'AUTHENTICATION_AUTHORITY_ASSERTED_IDENTITY' : 'S-1-18-1',
	'SERVICE_ASSERTED_IDENTITY' : 'S-1-18-2',
	'FRESH_PUBLIC_KEY_IDENTITY' : 'S-1-18-3',
	'KEY_TRUST_IDENTITY' : 'S-1-18-4',
	'KEY_PROPERTY_MFA' : 'S-1-18-5',
	'KEY_PROPERTY_ATTESTATION' : 'S-1-18-6',
}

well_known_sids_sid_name_map = {v: k for k, v in well_known_sids_name_sid_map.items()}

# below is the list of well-known SIDs which needs RE parsing for lookup
well_known_sids_re_name_sid_map = {
	'LOGON_ID' : 'S-1-5-5-x-y',
	'ENTERPRISE_READONLY_DOMAIN_CONTROLLERS' : 'S-1-5-21-<root domain>-498',
	'ADMINISTRATOR' : 'S-1-5-21-<machine>-500',
	'GUEST' : 'S-1-5-21-<machine>-501',
	'KRBTG' : 'S-1-5-21-<domain>-502',
	'DOMAIN_ADMINS' : 'S-1-5-21-<domain>-512',
	'DOMAIN_USERS' : 'S-1-5-21-<domain>-513',
	'DOMAIN_GUESTS' : 'S-1-5-21-<domain>-514',
	'DOMAIN_COMPUTERS' : 'S-1-5-21-<domain>-515',
	'DOMAIN_DOMAIN_CONTROLLERS' : 'S-1-5-21-<domain>-516',
	'CERT_PUBLISHERS' : 'S-1-5-21-<domain>-517',
	'SCHEMA_ADMINISTRATORS' : 'S-1-5-21-<root-domain>-518',
	'ENTERPRISE_ADMINS' : 'S-1-5-21-<root-domain>-519',
	'GROUP_POLICY_CREATOR_OWNERS' : 'S-1-5-21-<domain>-520',
	'READONLY_DOMAIN_CONTROLLERS' : 'S-1-5-21-<domain>-521',
	'CLONEABLE_CONTROLLERS' : 'S-1-5-21-<domain>-522',
	'PROTECTED_USERS' : 'S-1-5-21-<domain>-525',
	'KEY_ADMINS' : 'S-1-5-21-<domain>-526',
	'ENTERPRISE_KEY_ADMINS' : 'S-1-5-21-<domain>-527',
	'RAS_SERVERS' : 'S-1-5-21-<domain>-553',
	'ALLOWED_RODC_PASSWORD_REPLICATION_GROUP' : 'S-1-5-21-<domain>-571',
	'DENIED_RODC_PASSWORD_REPLICATION_GROUP' : 'S-1-5-21-<domain>-572',
}
	
well_known_sids_re_sid_name_map = {v: k for k, v in well_known_sids_re_name_sid_map.items()}


class DOMAIN_ALIAS_RID(enum.Enum):
	ADMINS = 0x00000220 # 	A local group used for administration of the domain.
	USERS = 0x00000221 #A local group that represents all users in the domain.
	GUESTS = 0x00000222 #A local group that represents guests of the domain.
	POWER_USERS = 0x00000223 #A local group used to represent a user or set of users who expect to treat a system as if it were their personal computer rather than as a workstation for multiple users.
	ACCOUNT_OPS = 	0x00000224 #A local group that exists only on systems running server operating systems. This local group permits control over nonadministrator accounts.
	SYSTEM_OPS = 0x00000225 #A local group that exists only on systems running server operating systems. This local group performs system administrative functions, not including security functions. It establishes network shares, controls printers, unlocks workstations, and performs other operations.
	PRINT_OPS = 0x00000226 #A local group that exists only on systems running server operating systems. This local group controls printers and print queues.
	BACKUP_OPS = 0x00000227 #A local group used for controlling assignment of file backup-and-restore privileges.
	REPLICATOR = 0x00000228 #A local group responsible for copying security databases from the primary domain controller to the backup domain controllers. These accounts are used only by the system.
	RAS_SERVERS = 0x00000229 #A local group that represents RAS and IAS servers. This group permits access to various attributes of user objects.
	PREW2KCOMPACCESS = 0x0000022A #A local group that exists only on systems running Windows 2000 Server. For more information, see Allowing Anonymous Access.
	REMOTE_DESKTOP_USERS = 0x0000022B #A local group that represents all remote desktop users.
	NETWORK_CONFIGURATION_OPS = 0x0000022C #A local group that represents the network configuration.
	INCOMING_FOREST_TRUST_BUILDERS = 0x0000022D #A local group that represents any forest trust users.
	MONITORING_USERS = 0x0000022E #A local group that represents all users being monitored.
	LOGGING_USERS = 0x0000022F #A local group responsible for logging users.
	AUTHORIZATIONACCESS = 0x00000230 #A local group that represents all authorized access.
	TS_LICENSE_SERVERS = 0x00000231 #A local group that exists only on systems running server operating systems that allow for terminal services and remote access.
	DCOM_USERS = 0x00000232 #A local group that represents users who can use Distributed Component Object Model (DCOM).
	IUSERS = 0X00000238 #A local group that represents Internet users.
	CRYPTO_OPERATORS = 0x00000239 #A local group that represents access to cryptography operators.
	CACHEABLE_PRINCIPALS_GROUP = 0x0000023B #A local group that represents principals that can be cached.
	NON_CACHEABLE_PRINCIPALS_GROUP = 0x0000023C #A local group that represents principals that cannot be cached.
	EVENT_LOG_READERS_GROUP = 0x0000023D #A local group that represents event log readers.
	CERTSVC_DCOM_ACCESS_GROUP = 0x0000023E #The local group of users who can connect to certification authorities using Distributed Component Object Model (DCOM).
	RDS_REMOTE_ACCESS_SERVERS = 0x0000023F  #A local group that represents RDS remote access servers.
	RDS_ENDPOINT_SERVERS = 0x00000240 #A local group that represents endpoint servers.
	RDS_MANAGEMENT_SERVERS = 0x00000241 #A local group that represents management servers.
	HYPER_V_ADMINS = 0x00000242 #A local group that represents hyper-v admins
	ACCESS_CONTROL_ASSISTANCE_OPS = 0x00000243 #A local group that represents access control assistance OPS.
	REMOTE_MANAGEMENT_USERS = 0x00000244 #A local group that represents remote management users.
	DEFAULT_ACCOUNT = 0x00000245 #A local group that represents the default account.
	STORAGE_REPLICA_ADMINS = 0x00000246 #A local group that represents storage replica admins.
	DEVICE_OWNERS = 0x00000247 #A local group that represents can make settings expected for Device Owners.

class DOMAIN_GROUP_RID(enum.Enum):
	ADMINS = 0x00000200 #The domain administrators' group. This account exists only on systems running server operating systems.
	USERS = 0x00000201 #A group that contains all user accounts in a domain. All users are automatically added to this group.
	GUESTS = 0x00000202 #The guest-group account in a domain.
	COMPUTERS = 0x00000203 #The domain computers' group. All computers in the domain are members of this group.
	CONTROLLERS = 0x00000204 #The domain controllers' group. All DCs in the domain are members of this group.
	CERT_ADMINS = 0x00000205 #The certificate publishers' group. Computers running Certificate Services are members of this group.
	ENTERPRISE_READONLY_DOMAIN_CONTROLLERS = 0x000001F2 #The group of enterprise read-only domain controllers.
	SCHEMA_ADMINS = 0x00000206 #The schema administrators' group. Members of this group can modify the Active Directory schema.
	ENTERPRISE_ADMINS = 0x00000207 #The enterprise administrators' group. Members of this group have full access to all domains in the Active Directory forest. Enterprise administrators are responsible for forest-level operations such as adding or removing new domains.
	POLICY_ADMINS = 0x00000208 #The policy administrators' group.
	READONLY_CONTROLLERS = 0x00000209 #The group of read-only domain controllers.
	CLONEABLE_CONTROLLERS = 0x0000020A #The group of cloneable domain controllers.
	CDC_RESERVED = 0x0000020C #The reserved CDC group.
	PROTECTED_USERS = 0x0000020D #	The protected users group.
	KEY_ADMINS = 0x0000020E #The key admins group.
	ENTERPRISE_KEY_ADMINS = 0x0000020F

class SECURITY_MANDATORY(enum.Enum):
	UNTRUSTED_RID = 0x00000000 #Untrusted.
	LOW_RID = 0x00001000 #Low integrity.
	MEDIUM_RID = 0x00002000 #Medium integrity.
	MEDIUM_PLUS_RID = 0x00002000 + 0x100 #Medium high integrity.
	HIGH_RID = 0x00003000 #High integrity.
	SYSTEM_RID = 0x00004000 #System integrity.
	PROTECTED_PROCESS_RID = 0x00005000

DOMAIN_USER_RID_ADMIN = 0x000001F4
DOMAIN_USER_RID_GUEST = 0x000001F5

SECURITY_LOCAL_SERVICE_RID  = 0x00000013
SECURITY_SERVER_LOGON_RID = 9
SECURITY_NETWORK_SERVICE_RID = 0x00000014

# https://docs.microsoft.com/en-us/windows/win32/secauthz/sid-strings
sddl_name_val_map = {
	"AN" : "S-1-5-7", #Anonymous logon. The corresponding RID is SECURITY_ANONYMOUS_LOGON_RID.
	"AO" : 	DOMAIN_ALIAS_RID.ACCOUNT_OPS.value, #Account operators. The corresponding RID is DOMAIN_ALIAS_RID_ACCOUNT_OPS.
	"AU" : 	"S-1-5-11", #Authenticated users. The corresponding RID is SECURITY_AUTHENTICATED_USER_RID.
	"BA" : 	DOMAIN_ALIAS_RID.ADMINS.value, #Built-in administrators. The corresponding RID is DOMAIN_ALIAS_RID_ADMINS.
	"BG" :  DOMAIN_ALIAS_RID.GUESTS.value, #Built-in guests. The corresponding RID is DOMAIN_ALIAS_RID_GUESTS.
	"BO" : 	DOMAIN_ALIAS_RID.BACKUP_OPS.value, #Backup operators. The corresponding RID is DOMAIN_ALIAS_RID_BACKUP_OPS.
	"BU" : 	"S-1-5-32-545", #Built-in users. The corresponding RID is DOMAIN_ALIAS_RID_USERS.
	"CA" :  DOMAIN_GROUP_RID.CERT_ADMINS.value, #Certificate publishers. The corresponding RID is DOMAIN_GROUP_RID_CERT_ADMINS.
	"CD" :  DOMAIN_ALIAS_RID.CERTSVC_DCOM_ACCESS_GROUP.value, #Users who can connect to certification authorities using Distributed Component Object Model (DCOM). The corresponding RID is DOMAIN_ALIAS_RID_CERTSVC_DCOM_ACCESS_GROUP.
	"CG" : 	"S-1-3", #Creator group. The corresponding RID is SECURITY_CREATOR_GROUP_RID.
	"CO" :  "S-1-3-0", #Creator owner. The corresponding RID is SECURITY_CREATOR_OWNER_RID.
	"DA" : 	DOMAIN_GROUP_RID.ADMINS.value, #Domain administrators. The corresponding RID is DOMAIN_GROUP_RID_ADMINS.
	"DC" : 	DOMAIN_GROUP_RID.COMPUTERS.value, #Domain computers. The corresponding RID is DOMAIN_GROUP_RID_COMPUTERS.
	"DD" : 	DOMAIN_GROUP_RID.CONTROLLERS.value, #Domain controllers. The corresponding RID is DOMAIN_GROUP_RID_CONTROLLERS.
	"DG" : 	DOMAIN_GROUP_RID.GUESTS.value, #Domain guests. The corresponding RID is DOMAIN_GROUP_RID_GUESTS.
	"DU" : 	DOMAIN_GROUP_RID.USERS.value, #Domain users. The corresponding RID is DOMAIN_GROUP_RID_USERS.
	"EA" : 	DOMAIN_GROUP_RID.ENTERPRISE_ADMINS.value, #Enterprise administrators. The corresponding RID is DOMAIN_GROUP_RID_ENTERPRISE_ADMINS.
	"ED" : 	SECURITY_SERVER_LOGON_RID, #Enterprise domain controllers. The corresponding RID is SECURITY_SERVER_LOGON_RID.
	"HI" : 	SECURITY_MANDATORY.HIGH_RID.value, #High integrity level. The corresponding RID is SECURITY_MANDATORY_HIGH_RID.
	"IU" : 	"S-1-5-4", #Interactively logged-on user. This is a group identifier added to the token of a process when it was logged on interactively. The corresponding logon type is LOGON32_LOGON_INTERACTIVE. The corresponding RID is SECURITY_INTERACTIVE_RID.
	"LA" : 	DOMAIN_USER_RID_ADMIN, #Local administrator. The corresponding RID is DOMAIN_USER_RID_ADMIN.
	"LG" : 	DOMAIN_USER_RID_GUEST, #Local guest. The corresponding RID is DOMAIN_USER_RID_GUEST.
	"LS" :  SECURITY_LOCAL_SERVICE_RID, #Local service account. The corresponding RID is SECURITY_LOCAL_SERVICE_RID.
	"LW" : 	SECURITY_MANDATORY.LOW_RID.value, #Low integrity level. The corresponding RID is SECURITY_MANDATORY_LOW_RID.
	"ME" : 	SECURITY_MANDATORY.MEDIUM_RID.value, #Medium integrity level. The corresponding RID is SECURITY_MANDATORY_MEDIUM_RID.
	# TODO ERROR: NO VALUE FOUND FOR THIS! "MU" :  SDDL_PERFMON_USERS, #Performance Monitor users.
	"NO" : 	DOMAIN_ALIAS_RID.NETWORK_CONFIGURATION_OPS.value, #Network configuration operators. The corresponding RID is DOMAIN_ALIAS_RID_NETWORK_CONFIGURATION_OPS.
	"NS" : 	SECURITY_NETWORK_SERVICE_RID, #Network service account. The corresponding RID is SECURITY_NETWORK_SERVICE_RID.
	"NU" : 	"S-1-5-2", #Network logon user. This is a group identifier added to the token of a process when it was logged on across a network. The corresponding logon type is LOGON32_LOGON_NETWORK. The corresponding RID is SECURITY_NETWORK_RID.
	"PA" : 	DOMAIN_GROUP_RID.POLICY_ADMINS.value, #Group Policy administrators. The corresponding RID is DOMAIN_GROUP_RID_POLICY_ADMINS.
	"PO" : 	DOMAIN_ALIAS_RID.PRINT_OPS.value, #Printer operators. The corresponding RID is DOMAIN_ALIAS_RID_PRINT_OPS.
	"PS" : 	"S-1-5-10", #Principal self. The corresponding RID is SECURITY_PRINCIPAL_SELF_RID.
	"PU" : 	DOMAIN_ALIAS_RID.POWER_USERS.value, #Power users. The corresponding RID is DOMAIN_ALIAS_RID_POWER_USERS.
	"RC" : 	"S-1-5-12", #Restricted code. This is a restricted token created using the CreateRestrictedToken function. The corresponding RID is SECURITY_RESTRICTED_CODE_RID.
	"RD" : 	DOMAIN_ALIAS_RID.REMOTE_DESKTOP_USERS.value, #Terminal server users. The corresponding RID is DOMAIN_ALIAS_RID_REMOTE_DESKTOP_USERS.
	"RE" : 	DOMAIN_ALIAS_RID.REPLICATOR.value, #Replicator. The corresponding RID is DOMAIN_ALIAS_RID_REPLICATOR.
	"RO" : 	DOMAIN_GROUP_RID.ENTERPRISE_READONLY_DOMAIN_CONTROLLERS.value, #Enterprise Read-only domain controllers. The corresponding RID is DOMAIN_GROUP_RID_ENTERPRISE_READONLY_DOMAIN_CONTROLLERS.
	"RS" :  DOMAIN_ALIAS_RID.RAS_SERVERS.value, #RAS servers group. The corresponding RID is DOMAIN_ALIAS_RID_RAS_SERVERS.
	"RU" :	DOMAIN_ALIAS_RID.PREW2KCOMPACCESS.value, #Alias to grant permissions to accounts that use applications compatible with operating systems previous to Windows 2000. The corresponding RID is DOMAIN_ALIAS_RID_PREW2KCOMPACCESS.
	"SA" : 	DOMAIN_GROUP_RID.SCHEMA_ADMINS.value, #Schema administrators. The corresponding RID is DOMAIN_GROUP_RID_SCHEMA_ADMINS.
	"SI" : 	SECURITY_MANDATORY.SYSTEM_RID.value, #System integrity level. The corresponding RID is SECURITY_MANDATORY_SYSTEM_RID.
	"SO" : 	DOMAIN_ALIAS_RID.SYSTEM_OPS.value, #Server operators. The corresponding RID is DOMAIN_ALIAS_RID_SYSTEM_OPS.
	"SU" : 	"S-1-5-6", #Service logon user. This is a group identifier added to the token of a process when it was logged as a service. The corresponding logon type is LOGON32_LOGON_SERVICE. The corresponding RID is SECURITY_SERVICE_RID.
	"SY" :	"S-1-5-18", #Local system. The corresponding RID is SECURITY_LOCAL_SYSTEM_RID.
	"WD" : 	"S-1-1-0",
}

sddl_val_name_map = {v: k for k, v in sddl_name_val_map.items()}

if __name__ == '__main__':
	sid = SID.from_string('S-1-15-2-1')
	print(sid.to_sddl())



import winreg


def get_nameserver_candidates():
	"""
	Returns a list of IP addresses which potentially an active DNS server used by the current machine.
	Please be aware that it's not at all precise!
	"""
	keynames = ['NameServer', 'DhcpNameServer']
	def get_ips(iface, keyname):
		try:
			result = {}
			res, _ = winreg.QueryValueEx(iface,keyname)
			if len(res) > 0:
				if res.find(',') != -1:
					for x in res.split(','):
						result[x] = 0
				else:
					result[res] = 0
		except Exception as e:
			pass #print(e)
		finally:
			return list(result.keys())

	ips = {}
	hklm = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
	key = 'System\CurrentControlSet\Services\Tcpip\Parameters\Interfaces'
	aKey = winreg.OpenKey(hklm, key)
	for i in range(1024):
		try:
			asubkey_name=winreg.EnumKey(aKey,i)
			iface = winreg.OpenKey(aKey, asubkey_name)
			for keyname in keynames:
				for res in get_ips(iface, keyname):
					ips[res] = 0
		
		except OSError as e:
			if e.args[0] == 22:
				break
			elif e.args[0] == 259:
				continue
			else:
				raise
		except Exception as e:
			print(e)
	
	return list(ips.keys())

if __name__ == '__main__':
	x = get_nameserver_candidates()
	print(x)


import io
from winacl.dtyp.wcee.rsaprivkey import RSAPrivateKeyBlob


# https://github.com/openssl/openssl/blob/9dddcd90a1350fa63486cbf3226c3eee79f9aff5/crypto/pem/pvkfmt.c
class PVKFile:
	def __init__(self):
		self.magic:int = 0xb0b5f11e
		self.reserved:int = 0
		self.keytype:bytes = None
		self.isencrypted:int = None
		self.saltlength:int = None
		self.keylength:int = None
		self.saltblob:bytes = None
		self.keyblob:bytes = None

	@staticmethod
	def construct_unencrypted(keytype, keydata):
		if isinstance(keytype, str):
			keytype = keytype.encode()
		pvk = PVKFile()
		pvk.keytype = keytype
		pvk.isencrypted = 0
		pvk.saltlength = 0
		pvk.keylength = len(keydata)
		pvk.saltblob = b''
		pvk.keyblob = keydata
		return pvk

	@staticmethod
	def from_file(filepath):
		with open(filepath, 'rb') as f:
			return PVKFile.from_buffer(f)
	
	
	@staticmethod
	def from_bytes(data):
		return PVKFile.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		pvk = PVKFile()
		pvk.magic = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		assert 0xb0b5f11e == pvk.magic
		pvk.reserved = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pvk.keytype = buff.read(4)
		pvk.isencrypted = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pvk.saltlength = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pvk.keylength = int.from_bytes(buff.read(4), byteorder='little', signed=False)
		pvk.saltblob = buff.read(pvk.saltlength)
		pvk.keyblob = buff.read(pvk.keylength)
		return pvk

	def to_bytes(self):
		data = b''
		data += self.magic.to_bytes(4, byteorder='little', signed=False)
		data += self.reserved.to_bytes(4, byteorder='little', signed=False)
		data += self.keytype
		data += self.isencrypted.to_bytes(4, byteorder='little', signed=False)
		if self.saltblob is None:
			self.saltblob = b''
		if self.keyblob is None:
			self.keyblob = b''
		data += len(self.saltblob).to_bytes(4, byteorder='little', signed=False)
		data += len(self.keyblob).to_bytes(4, byteorder='little', signed=False)
		data += self.saltblob
		data += self.keyblob
		return data

	def get_key(self):
		if self.keyblob[8:].startswith(b'RSA2'):
			return RSAPrivateKeyBlob.from_bytes(self.keyblob).get_key()

	def __str__(self):
		t = ''
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
		return t


import ctypes
from winacl.functions.constants import SE_OBJECT_TYPE
from winacl.functions.kernel32 import CloseHandle, CreateFileW, GENERIC_READ, READ_CONTROL, OPEN_EXISTING, FILE_ATTRIBUTE_DIRECTORY, FILE_FLAG_BACKUP_SEMANTICS
from winacl.functions.advapi32 import LookupAccountNameW, LookupAccountSidW, \
	GetSecurityInfo, OWNER_SECURITY_INFORMATION, GROUP_SECURITY_INFORMATION, \
	DACL_SECURITY_INFORMATION, OpenSCManagerW, SC_MANAGER_ENUMERATE_SERVICE, \
	EnumServicesStatusW, OpenServiceW, QueryServiceObjectSecurity, CloseServiceHandle, \
	hive_name_map, RegOpenKeyExW, RegCloseKey, RegEnumKeyExW, RegEnumValueW, \
	SetSecurityInfo, BuildTrusteeWithSidW, GetEffectiveRightsFromAclW, \
	ConvertSidToStringSidW, ConvertStringSidToSidW, \
	ConvertSecurityDescriptorToStringSecurityDescriptorW, OpenProcessToken, GetTokenInformation, \
	LsaGetLogonSessionData, TokenStatistics, OpenProcess, GetCurrentProcessId
from winacl.functions.netapi32 import NetUserGetLocalGroups, NetUserGetInfo
import glob
import os
from winacl.dtyp.security_descriptor import SECURITY_DESCRIPTOR
from winacl.dtyp.ace import ACCESS_ALLOWED_ACE, AceFlags, FILE_ACCESS_MASK, ACCESS_MASK
from winacl.dtyp.sid import SID
from winacl.functions.rights_calc import EvaluateSidAgainstDescriptor

def get_sid_for_user(username):
	sid, domain, use = LookupAccountNameW(None, username)
	return sid

def get_user_for_sid(sid_str):
	sid = SID.from_string(sid_str)
	username, domain, use = LookupAccountSidW(None, sid.to_bytes())
	return '%s\\%s' % (domain, username)

def get_reg_sd(key_handle):
	sd = GetSecurityInfo(key_handle, SE_OBJECT_TYPE.SE_REGISTRY_KEY.value, OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION)
	return sd

def set_file_sd(file_path, sd):
	req_rights = 0x10000000	#GENERIC_ALL
	file_handle = CreateFileW(file_path, dwDesiredAccess = req_rights, dwCreationDisposition = OPEN_EXISTING)
	SetSecurityInfo(file_handle, SE_OBJECT_TYPE.SE_FILE_OBJECT.value, sd)
	CloseHandle(file_handle)

def get_file_sd(file_path):
	file_handle = CreateFileW(file_path, dwDesiredAccess = READ_CONTROL, dwCreationDisposition = OPEN_EXISTING)
	sd = GetSecurityInfo(file_handle, SE_OBJECT_TYPE.SE_FILE_OBJECT.value, OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION)
	CloseHandle(file_handle)
	return sd

def get_directory_sd(dir_path):
	file_handle = CreateFileW(dir_path, dwDesiredAccess = READ_CONTROL, dwCreationDisposition = OPEN_EXISTING, dwFlagsAndAttributes = FILE_ATTRIBUTE_DIRECTORY | FILE_FLAG_BACKUP_SEMANTICS)
	sd = GetSecurityInfo(file_handle, SE_OBJECT_TYPE.SE_FILE_OBJECT.value, OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION)
	CloseHandle(file_handle)
	return sd

def get_dir_file_recursive(dir_path, with_files = False):
	if dir_path[-1] != '\\':
		dir_path += '\\'
	for filename in glob.iglob(dir_path + '**/*', recursive=True):
		if os.path.isdir(filename) is True:
			try:
				yield filename, 'dir', get_directory_sd(filename)
			except:
				yield filename, 'dir', 'ERR'
		elif with_files is True and os.path.isfile(filename) is True:
			try:
				yield filename, 'file', get_file_sd(filename)
			except:
				yield filename, 'file', 'ERR'

def get_servicemanager_sd():
	scm_handle = OpenSCManagerW(dwDesiredAccess = READ_CONTROL )
	sd = QueryServiceObjectSecurity(scm_handle, dwSecurityInformation = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION, sd_object_type = SE_OBJECT_TYPE.SE_SERVICE)
	CloseServiceHandle(scm_handle)
	return sd

def get_service_sd(service_name):
	scm_handle = OpenSCManagerW(dwDesiredAccess = READ_CONTROL | SC_MANAGER_ENUMERATE_SERVICE )
	service_handle = OpenServiceW(scm_handle, service_name, dwDesiredAccess = READ_CONTROL)
	sd = QueryServiceObjectSecurity(service_handle, dwSecurityInformation = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION, sd_object_type = SE_OBJECT_TYPE.SE_SERVICE)
	CloseServiceHandle(scm_handle)
	CloseServiceHandle(service_handle)
	return sd

def enumerate_all_service_sd():
	yield 'SCM', get_servicemanager_sd()
	scm_handle = OpenSCManagerW(dwDesiredAccess = READ_CONTROL | SC_MANAGER_ENUMERATE_SERVICE )
	for service_name in EnumServicesStatusW(scm_handle):
		try:
			service_handle = OpenServiceW(scm_handle, service_name, dwDesiredAccess = READ_CONTROL)
			sd = QueryServiceObjectSecurity(service_handle, dwSecurityInformation = OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION, sd_object_type = SE_OBJECT_TYPE.SE_SERVICE)
			yield service_name, sd
		except Exception as e:
			yield service_name, 'err'
		finally:
			try:
				CloseServiceHandle(service_handle)
			except:
				pass
	try:
		CloseServiceHandle(scm_handle)
	except:
		pass

def get_registry_hive_sd(hive_name):
	key_handle = hive_name_map[hive_name]
	sd = GetSecurityInfo(key_handle, SE_OBJECT_TYPE.SE_REGISTRY_KEY.value, OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION)
	return sd

def get_registry_key_sd(reg_path):
	# eg. HKLM\\SAM
	path_elements = reg_path.split('\\')
	reg_handle = hive_name_map[path_elements[0]]
	handles = []
	for name in path_elements[1:-1]:
		reg_handle = RegOpenKeyExW(reg_handle, name)
		handles.append(reg_handle)
	
	sd = GetSecurityInfo(reg_handle, SE_OBJECT_TYPE.SE_REGISTRY_KEY.value, OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION)
	for reg_handle in handles[::-1]:
		RegCloseKey(reg_handle)
	
	return sd

#def get_maximum_permissions_for_sid(sd, sid):
#	EvaluateSidAgainstDescriptor

def get_maximum_permissions_for_user(sd, username):
	sid_groups = []
	domain = None
	if username.find('\\') != -1:
		domain, username = username.split('\\')

	user_sid, *t = LookupAccountNameW(domain, username)
	#print(str(user_sid))
	sid_groups.append(user_sid)
	for group_name in NetUserGetLocalGroups(domain, username):
		#print(group_name)
		sid, groupname, use = LookupAccountNameW(domain, group_name)
		#print(str(sid))
		sid_groups.append(sid)
	#print(win32net.NetUserGetLocalGroups('TEST', 'victim'))
	#input([str(x) for x in sid_groups])
	res, mask = EvaluateSidAgainstDescriptor(sd, user_sid, 0x02000000, sid_groups)
	return mask

def enumerate_registry_sd(reg_path, reg_handle = None):
	# TODO: do this
	if reg_path[-1] == '\\':
		reg_path = reg_path[:-1]
	handles = []
	if reg_handle is None:
		path_elements = reg_path.split('\\')
		reg_handle = hive_name_map[path_elements[0]]
		handles = []
		for name in path_elements[1:]:
			reg_handle = RegOpenKeyExW(reg_handle, name)
			handles.append(reg_handle)

	subkeys = RegEnumKeyExW(reg_handle)
	#print(subkeys)
	for sk in subkeys:
		reg_path_sk = '%s\\%s' % (reg_path, sk)
		#print(reg_path_sk)
		try:
			reg_handle_sk = RegOpenKeyExW(reg_handle, sk)
		except Exception as e:
			if str(e).find('cannot find') != -1:
				print(str(e) + reg_path_sk)
			else:
				print(e)
			
		else:
			sd = GetSecurityInfo(reg_handle, SE_OBJECT_TYPE.SE_REGISTRY_KEY.value, OWNER_SECURITY_INFORMATION | GROUP_SECURITY_INFORMATION | DACL_SECURITY_INFORMATION, se_object_type = SE_OBJECT_TYPE.SE_REGISTRY_KEY)
			yield (reg_path_sk, sd)
			yield from enumerate_registry_sd(reg_path_sk, reg_handle_sk)

	for reg_handle in handles:
		RegCloseKey(reg_handle)

def get_logon_info():
	phandle = OpenProcess(None , False, GetCurrentProcessId())
	tokhandle = OpenProcessToken(phandle)
	stats = GetTokenInformation(tokhandle, TokenStatistics)
	return LsaGetLogonSessionData(stats.AuthenticationId)
